"""Wrapper for Notion API blocks.

Blocks are the base for all Notion content.

For validation the Pydantic model fields specify if a field is optional or not.
Some fields are always set, e.g. `id`, when retrieving an object but must not be set
when sending the object to the Notion API in order to create the object.
To model this behavior, the default sentinel value `Unset` is used for those objects, e.g.
```
class SelectOption(GenericObject)
    id: str | UnsetType = Unset
```
Be aware that this is important when updating to differentiate between the actual set
values from default/unset values.
"""

from __future__ import annotations

from abc import ABC
from typing import Annotated, Any, cast

from pydantic import AfterValidator, Field, SerializeAsAny
from typing_extensions import TypeVar

from ultimate_notion.obj_api.core import GenericObject, NotionEntity, TypedObject, Unset, UnsetType
from ultimate_notion.obj_api.enums import BGColor, CodeLang, Color
from ultimate_notion.obj_api.objects import (
    MAX_TEXT_OBJECT_SIZE,
    Annotations,
    BlockRef,
    CustomEmojiObject,
    EmojiObject,
    FileObject,
    MentionDatabase,
    MentionMixin,
    MentionObject,
    MentionPage,
    ParentRef,
    RichTextBaseObject,
    TextObject,
    UserRef,
)
from ultimate_notion.obj_api.props import PropertyValue, Title
from ultimate_notion.obj_api.schema import Property


class DataObject(NotionEntity, ABC):
    """The base type for all Notion objects that hold actual data."""

    has_children: bool = False

    in_trash: bool = False  # used to be `archived`
    archived: bool = False  # ToDo: Deprecated but still partially used in Notion. Check to remove in v1.0!

    last_edited_by: UserRef | UnsetType = Unset


class Database(DataObject, MentionMixin, object='database'):
    """A database record type."""

    title: list[SerializeAsAny[RichTextBaseObject]] | UnsetType = Unset
    url: str | UnsetType = Unset
    public_url: str | None = None
    icon: SerializeAsAny[FileObject] | EmojiObject | CustomEmojiObject | None = None
    cover: SerializeAsAny[FileObject] | None = None
    properties: dict[str, SerializeAsAny[Property]]
    description: list[SerializeAsAny[RichTextBaseObject]] | UnsetType = Unset
    is_inline: bool = False

    def build_mention(self, style: Annotations | None = None) -> MentionObject:
        return MentionDatabase.build_mention_from(self, style=style)


class Page(DataObject, MentionMixin, object='page'):
    """A standard Notion page object."""

    url: str | UnsetType = Unset
    public_url: str | None = None
    icon: SerializeAsAny[FileObject] | EmojiObject | CustomEmojiObject | None = None
    cover: SerializeAsAny[FileObject] | None = None
    properties: dict[str, PropertyValue]
    is_locked: bool | UnsetType = Unset

    def _get_title_prop_name(self) -> str:
        """Get the name of the title property."""
        # As the 'title' property might be renamed in case of pages in databases, we look for the `id`.
        for name, prop in self.properties.items():
            if prop.id == 'title':
                return name
        msg = 'Encountered a page without title property'
        raise RuntimeError(msg)

    @property
    def title(self) -> list[RichTextBaseObject]:
        """Retrieve the title of the page from page properties."""
        title_prop_name = self._get_title_prop_name()
        title_prop = cast(Title, self.properties[title_prop_name])
        return title_prop.title

    def build_mention(self, style: Annotations | None = None) -> MentionObject:
        return MentionPage.build_mention_from(self, style=style)


# ToDo: Use new syntax when requires-python >= 3.12
GO_co = TypeVar('GO_co', bound=GenericObject, default=GenericObject, covariant=True)


class Block(TypedObject[GO_co], DataObject, object='block', polymorphic_base=True):
    """A standard block object in Notion.

    !!! note
        The equality and hash implementation only considers the content of the block,
        i.e., the `type` and the actual type data, but ignores meta fields like
        `id`, `created_time`, `last_edited_time`, etc.
    """

    def __hash__(self) -> int:
        return hash((self.type, self.value))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Block):
            return NotImplemented

        # ignore meta fields, e.g. id, created_time, etc. for equality, only use content
        return self.type == other.type and self.value == other.value


class UnsupportedBlockTypeData(GenericObject):
    """Type data for `UnsupportedBlock`."""


class UnsupportedBlock(Block[UnsupportedBlockTypeData], type='unsupported'):
    """A placeholder for unsupported blocks in the API."""

    unsupported: UnsupportedBlockTypeData = Field(default_factory=UnsupportedBlockTypeData)


def normalize_text_objs(rt_objs: list[RichTextBaseObject]) -> list[RichTextBaseObject]:
    # When sending text objects to the Notion API the number of characters is limited to MAX_TEXT_OBJECT_SIZE,
    # as defined under https://developers.notion.com/reference/request-limits#limits-for-property-values.
    # The API response does not enforce this limit, though. We therefore chunk the text here to allow comparison.
    result: list[RichTextBaseObject] = []
    for rt_obj in rt_objs:
        if isinstance(rt_obj, TextObject) and len(rt_obj.text.content) > MAX_TEXT_OBJECT_SIZE:
            for i in range(0, len(rt_obj.text.content), MAX_TEXT_OBJECT_SIZE):
                content_chunk = rt_obj.text.content[i : i + MAX_TEXT_OBJECT_SIZE]
                plain_text_chunk = rt_obj.plain_text[i : i + MAX_TEXT_OBJECT_SIZE]
                new_rt_obj = rt_obj.model_copy(
                    update={
                        'text': rt_obj.text.model_copy(update={'content': content_chunk}),
                        'plain_text': plain_text_chunk,
                    }
                )
                result.append(new_rt_obj)
        else:
            result.append(rt_obj)
    return result


class TextBlockTypeData(GenericObject):
    """Type data for `TextBlock`."""

    rich_text: Annotated[
        list[SerializeAsAny[RichTextBaseObject]], AfterValidator(normalize_text_objs), Field(default_factory=list)
    ]


# ToDo: Use new syntax when requires-python >= 3.12
TB_co = TypeVar('TB_co', bound=TextBlockTypeData, default=TextBlockTypeData, covariant=True)


class TextBlock(Block[TB_co]):
    """A standard abstract text block object in Notion."""


class ColoredTextBlockTypeData(TextBlockTypeData):
    """Type data for `TextBlock` with color."""

    color: Color | BGColor = Color.DEFAULT


# ToDo: Use new syntax when requires-python >= 3.12
CTB_co = TypeVar('CTB_co', bound=ColoredTextBlockTypeData, default=ColoredTextBlockTypeData, covariant=True)


class ColoredTextBlock(TextBlock[CTB_co]):
    """A standard abstract text block object in Notion with color."""


class ParagraphTypeData(ColoredTextBlockTypeData):
    """Type data for `Paragraph` block."""

    children: list[SerializeAsAny[Block]] = Field(default_factory=list)


class Paragraph(ColoredTextBlock[ParagraphTypeData], type='paragraph'):
    """A paragraph block in Notion."""

    paragraph: ParagraphTypeData = Field(default_factory=ParagraphTypeData)  # type: ignore[arg-type]


class HeadingTypeData(ColoredTextBlockTypeData):
    """Type data for `Heading` block."""

    is_toggleable: bool = False


class Heading(ColoredTextBlock[HeadingTypeData]):
    """Abstract Heading block."""


class Heading1(Heading, type='heading_1'):
    """A heading_1 block in Notion."""

    heading_1: HeadingTypeData = Field(default_factory=HeadingTypeData)  # type: ignore[arg-type]


class Heading2(Heading, type='heading_2'):
    """A heading_2 block in Notion."""

    heading_2: HeadingTypeData = Field(default_factory=HeadingTypeData)  # type: ignore[arg-type]


class Heading3(Heading, type='heading_3'):
    """A heading_3 block in Notion."""

    heading_3: HeadingTypeData = Field(default_factory=HeadingTypeData)  # type: ignore[arg-type]


class QuoteTypeData(ColoredTextBlockTypeData):
    """Type data for `Quote` block."""

    children: list[SerializeAsAny[Block]] = Field(default_factory=list)


class Quote(ColoredTextBlock[QuoteTypeData], type='quote'):
    """A quote block in Notion."""

    quote: QuoteTypeData = Field(default_factory=QuoteTypeData)  # type: ignore[arg-type]


class CaptionMixin(GenericObject, ABC):
    """Mixin for blocks having a caption."""

    caption: list[SerializeAsAny[RichTextBaseObject]] = Field(default_factory=list)


class CodeTypeData(TextBlockTypeData, CaptionMixin):
    """Type data for `Code` block."""

    language: CodeLang = CodeLang.PLAIN_TEXT


class Code(TextBlock[CodeTypeData], type='code'):
    """A code block in Notion."""

    code: CodeTypeData = Field(default_factory=CodeTypeData)  # type: ignore[arg-type]


class CalloutTypeData(ColoredTextBlockTypeData):
    """Type data for `Callout` block."""

    # `children` is undocumented and behaves inconsistent. It is used during creation but not filled when retrieved.
    children: list[SerializeAsAny[Block]] = Field(default_factory=list)
    icon: SerializeAsAny[FileObject] | EmojiObject | CustomEmojiObject | UnsetType = Unset


class Callout(ColoredTextBlock[CalloutTypeData], type='callout'):
    """A callout block in Notion."""

    callout: CalloutTypeData = Field(default_factory=CalloutTypeData)  # type: ignore[arg-type]


class BulletedListItemTypeData(ColoredTextBlockTypeData):
    """Type data for `BulletedListItem` block."""

    children: list[SerializeAsAny[Block]] = Field(default_factory=list)


class BulletedListItem(ColoredTextBlock[BulletedListItemTypeData], type='bulleted_list_item'):
    """A bulleted list item in Notion."""

    bulleted_list_item: BulletedListItemTypeData = Field(default_factory=BulletedListItemTypeData)  # type: ignore[arg-type]


class NumberedListItemTypeData(ColoredTextBlockTypeData):
    """Type data for `NumberedListItem` block."""

    children: list[SerializeAsAny[Block]] = Field(default_factory=list)


class NumberedListItem(ColoredTextBlock[NumberedListItemTypeData], type='numbered_list_item'):
    """A numbered list item in Notion."""

    numbered_list_item: NumberedListItemTypeData = Field(default_factory=NumberedListItemTypeData)  # type: ignore[arg-type]


class ToDoTypeData(ColoredTextBlockTypeData):
    """Type data for `ToDo` block."""

    checked: bool = False
    children: list[SerializeAsAny[Block]] = Field(default_factory=list)


class ToDo(ColoredTextBlock[ToDoTypeData], type='to_do'):
    """A todo list item in Notion."""

    to_do: ToDoTypeData = Field(default_factory=ToDoTypeData)  # type: ignore[arg-type]


class ToggleTypeData(ColoredTextBlockTypeData):
    """Type data for `Toggle` block."""

    children: list[SerializeAsAny[Block]] = Field(default_factory=list)


class Toggle(ColoredTextBlock[ToggleTypeData], type='toggle'):
    """A toggle list item in Notion."""

    toggle: ToggleTypeData = Field(default_factory=ToggleTypeData)  # type: ignore[arg-type]


class DividerTypeData(GenericObject):
    """Type data for `Divider` block."""


class Divider(Block[DividerTypeData], type='divider'):
    """A divider block in Notion."""

    divider: DividerTypeData = Field(default_factory=DividerTypeData)


class TableOfContentsTypeData(GenericObject):
    """Type data for `TableOfContents` block."""

    color: Color | BGColor = Color.DEFAULT


class TableOfContents(Block[TableOfContentsTypeData], type='table_of_contents'):
    """A table_of_contents block in Notion."""

    table_of_contents: TableOfContentsTypeData = Field(default_factory=TableOfContentsTypeData)


class BreadcrumbTypeData(GenericObject):
    """Type data for `Breadcrumb` block."""


class Breadcrumb(Block[BreadcrumbTypeData], type='breadcrumb'):
    """A breadcrumb block in Notion."""

    breadcrumb: BreadcrumbTypeData = Field(default_factory=BreadcrumbTypeData)


class EmbedTypeData(CaptionMixin):
    """Type data for `Embed` block."""

    url: str


class Embed(Block[EmbedTypeData], type='embed'):
    """An embed block in Notion."""

    embed: EmbedTypeData

    @classmethod
    def build(cls, url: str, caption: list[RichTextBaseObject]) -> Embed:
        return Embed.model_construct(embed=EmbedTypeData.model_construct(url=url, caption=caption))


class BookmarkTypeData(CaptionMixin):
    """Type data for `Bookmark` block."""

    url: str


class Bookmark(Block[BookmarkTypeData], type='bookmark'):
    """A bookmark block in Notion."""

    bookmark: BookmarkTypeData

    @classmethod
    def build(cls, url: str, caption: list[RichTextBaseObject]) -> Bookmark:
        return Bookmark.model_construct(bookmark=BookmarkTypeData.model_construct(url=url, caption=caption))


class LinkPreviewTypeData(GenericObject):
    """Type data for `LinkPreview` block."""

    url: str


class LinkPreview(Block[LinkPreviewTypeData], type='link_preview'):
    """A link_preview block in Notion."""

    link_preview: LinkPreviewTypeData


class EquationTypeData(GenericObject):
    """Type data for `Equation` block."""

    expression: str | None = None


class Equation(Block[EquationTypeData], type='equation'):
    """An equation block in Notion."""

    equation: EquationTypeData = Field(default_factory=EquationTypeData)


class FileBase(Block[FileObject], ABC):
    """A abstract block referencing a FileObject."""


class File(FileBase, type='file'):
    """A file block in Notion."""

    file: SerializeAsAny[FileObject] | None = None


class Image(FileBase, type='image'):
    """An image block in Notion."""

    image: SerializeAsAny[FileObject] | None = None


class Video(FileBase, type='video'):
    """A video block in Notion."""

    video: SerializeAsAny[FileObject] | None = None


class PDF(FileBase, type='pdf'):
    """A pdf block in Notion."""

    pdf: SerializeAsAny[FileObject] | None = None


class Audio(FileBase, type='audio'):
    """An audio block in Notion."""

    audio: SerializeAsAny[FileObject] | None = None


class ChildPageTypeData(GenericObject):
    """Type data for `ChildPage` block."""

    title: str


class ChildPage(Block[ChildPageTypeData], type='child_page'):
    """A child page block in Notion."""

    child_page: ChildPageTypeData


class ChildDatabaseTypeData(GenericObject):
    """Type data for `ChildDatabase` block."""

    title: str


class ChildDatabase(Block[ChildDatabaseTypeData], type='child_database'):
    """A child database block in Notion."""

    child_database: ChildDatabaseTypeData


class ColumnTypeData(GenericObject):
    """Type data for `Column` block."""

    # note that children will not be populated when getting this block
    # https://developers.notion.com/changelog/column-list-and-column-support
    children: list[SerializeAsAny[Block]] = Field(default_factory=list)
    width_ratio: float | None = None


class Column(Block[ColumnTypeData], type='column'):
    """A column block in Notion."""

    column: ColumnTypeData = Field(default_factory=ColumnTypeData)

    @classmethod
    def build(cls, width_ratio: float | None = None) -> Column:
        return Column.model_construct(column=ColumnTypeData.model_construct(children=[], width_ratio=width_ratio))


class ColumnListTypeData(GenericObject):
    """Type data for `ColumnList` block."""

    # note that children will not be populated when getting this block
    # https://developers.notion.com/changelog/column-list-and-column-support
    children: list[Column] = Field(default_factory=list)


class ColumnList(Block[ColumnListTypeData], type='column_list'):
    """A column list block in Notion."""

    column_list: ColumnListTypeData = Field(default_factory=ColumnListTypeData)


class TableRowTypeData(GenericObject):
    """Type data for `TableRow` block."""

    cells: list[list[SerializeAsAny[RichTextBaseObject]]] = Field(default_factory=list)


class TableRow(Block[TableRowTypeData], type='table_row'):
    """A table_row block in Notion."""

    table_row: TableRowTypeData = Field(default_factory=TableRowTypeData)

    @classmethod
    def build(cls, n_cells: int) -> TableRow:
        return TableRow.model_construct(table_row=TableRowTypeData.model_construct(cells=[[] for _ in range(n_cells)]))


class TableTypeData(GenericObject):
    """Type data for `Table` block."""

    table_width: int = 0
    has_column_header: bool = False
    has_row_header: bool = False
    # note that children will not be populated when getting this block
    # https://developers.notion.com/reference/block#table-blocks
    children: list[TableRow] = Field(default_factory=list)


class Table(Block[TableTypeData], type='table'):
    """A table block in Notion."""

    table: TableTypeData = Field(default_factory=TableTypeData)


class LinkToPage(Block[ParentRef], type='link_to_page'):
    """A link_to_page block in Notion."""

    link_to_page: SerializeAsAny[ParentRef]


class SyncedBlockTypeData(GenericObject):
    """Type data for `SyncedBlock` block."""

    synced_from: BlockRef | None = None
    children: list[SerializeAsAny[Block]] = Field(default_factory=list)


class SyncedBlock(Block[SyncedBlockTypeData], type='synced_block'):
    """A synced_block block in Notion - either original or synced."""

    synced_block: SyncedBlockTypeData = Field(default_factory=SyncedBlockTypeData)

    def serialize_for_api(self) -> dict[str, Any]:
        """Serialize the object for sending it to the Notion API."""
        model_data = super().serialize_for_api()
        # Everywhere else we have to remove "null" values but `synced_from` is an exception
        if self.synced_block.synced_from is None:
            model_data['synced_block']['synced_from'] = None
        return model_data


class TemplateTypeData(TextBlockTypeData):
    """Type data for `Template` block."""

    children: list[SerializeAsAny[Block]] = Field(default_factory=list)


class Template(TextBlock[TemplateTypeData], type='template'):
    """A template block in Notion."""

    template: TemplateTypeData = Field(default_factory=TemplateTypeData)  # type: ignore[arg-type]



"""Functionality around defining a database schema.

Currently only normal databases, no wiki databases, can be created [1].
Neither the `Unique ID` nor `Status` nor the `Verfication` page property can be set as a database property
in a custom Schema when creating the database.

[1] https://developers.notion.com/docs/working-with-databases#wiki-databases


### Design Principles

A schema is a subclass of `Schema` that holds `Property` objects, which define the name and the type of the
property, e.g. `Text`, `Number`.

The source of truth is always the `obj_ref` and a `Property` holds only auxilliary
information if actually needed. Since the object references `obj_ref` must always point
to the actual `obj_api.blocks.Database.properties` value if the schema is bound to a database,
the method `_set_obj_refs` rewires this when a schema is used to create a database.
"""

from __future__ import annotations

from abc import ABC, ABCMeta
from collections import Counter
from collections.abc import Iterator
from copy import deepcopy
from functools import partial
from typing import TYPE_CHECKING, Annotated, Any, TypeAlias, cast, overload

from pydantic import BaseModel, ConfigDict, Field, PrivateAttr, create_model, field_validator
from tabulate import tabulate
from typing_extensions import Self, TypeVar

import ultimate_notion.obj_api.core as obj_core
import ultimate_notion.obj_api.schema as obj_schema
from ultimate_notion import rich_text
from ultimate_notion.core import Wrapper, get_active_session, get_repr
from ultimate_notion.errors import (
    EmptyListError,
    InvalidAPIUsageError,
    MultipleItemsError,
    PropertyError,
    RelationError,
    RollupError,
    SchemaError,
    SchemaNotBoundError,
    UnsetError,
)
from ultimate_notion.obj_api.core import Unset, UnsetType, is_unset, raise_unset
from ultimate_notion.obj_api.enums import OptionGroupType
from ultimate_notion.obj_api.schema import AggFunc, NumberFormat
from ultimate_notion.option import Option, OptionGroup, OptionNS, check_for_updates
from ultimate_notion.props import PropertyValue
from ultimate_notion.utils import SList, dict_diff_str, is_notebook

if TYPE_CHECKING:
    from ultimate_notion.database import Database
    from ultimate_notion.obj_api.core import UnsetType
    from ultimate_notion.page import Page

T = TypeVar('T')
PropertyGO: TypeAlias = obj_schema.Property[obj_core.GenericObject]
GO_co = TypeVar('GO_co', bound=PropertyGO, default=PropertyGO, covariant=True)


class Property(Wrapper[GO_co], ABC, wraps=PropertyGO):
    """Base class for Notion property objects.

    A property defines the name and type of a property in a database, e.g. number, date, text, etc.
    """

    """Reference to the low-level object representation of this property"""
    allowed_at_creation = True
    """If the Notion API allows to create a new database with a property of this type"""
    _name: str | None = None  # name given by the user, not the Notion API, will match when set
    _owner: type[Schema] | None = None  # back reference to the schema
    _attr_name: str | None = None  # name of the attribute within the schema holding the Property

    def __new__(cls, *args: Any, **kwargs: Any) -> Property:
        if cls is Property:
            msg = f'{cls.__name__} is abstract and cannot be instantiated directly'
            raise TypeError(msg)
        return super().__new__(cls)

    def __init__(self, name: str | None = None, **kwargs: Any) -> None:
        if name is not None and not isinstance(name, str):
            msg = f'The name of the property must be a string, not {type(name).__name__}'
            raise PropertyError(msg)
        self._name = name
        obj_api_type = self._obj_api_map_inv[self.__class__]
        self.obj_ref = obj_api_type.build(**kwargs)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Property):
            return NotImplemented
        other_obj_ref = cast(GO_co, other.obj_ref)
        return (self.obj_ref.type == other_obj_ref.type) and (self.obj_ref.value == other_obj_ref.value)

    def __hash__(self) -> int:
        return hash((self.obj_ref.type, self.obj_ref.value))

    def __repr__(self) -> str:
        return get_repr(self, name=f'{self.__class__.__name__} Property', desc=self.name)

    def __str__(self) -> str:
        return self.__class__.__name__

    @property
    def _is_init(self) -> bool:
        """Determines if the property is already initialized"""
        return hasattr(self, '_obj_ref')

    @property
    def _is_init_ready(self) -> bool:
        """Determines if the property is ready to be initialized"""
        return True

    def _get_owner(self) -> type[Schema]:
        """Get the owner schema of this property."""
        if self._owner is None:
            msg = f'The property {self.name} is not bound to a Schema!'
            raise PropertyError(msg)
        return self._owner

    def _update_prop(self, prop_obj: GO_co) -> GO_co:  # type: ignore[misc] # breaking covariance
        """Update the attributes of this property from a schema."""
        db = self._get_owner().get_db()
        session = get_active_session()
        session.api.databases.update(db=db.obj_ref, schema={self.name: prop_obj})
        return cast(GO_co, db.obj_ref.properties[self.name])

    def _rename_prop(self, new_name: str | None) -> None:
        """Update the name of this property in the schema or delete it.

        !!! warning

            The name of the Python attribute, i.e. `attr_name`, in the schema is not the changed!
            Use `prop.attr_name = "..."` to change the attribute name of a property `prop` within a schema.
        """
        schema = self._get_owner()
        db = schema.get_db()
        session = get_active_session()
        schema_obj: dict[str, obj_schema.RenameProp | None] = (
            {self.name: None} if new_name is None else {self.name: obj_schema.RenameProp(name=new_name)}
        )
        session.api.databases.update(db=db.obj_ref, schema=schema_obj)

        if new_name is None:
            schema._props = [prop for prop in schema.get_props() if prop.name != self.name]
        else:
            self.obj_ref.name = new_name
            self._name = new_name

        schema._set_obj_refs()

    @classmethod
    def wrap_obj_ref(cls, obj_ref: GO_co) -> Self:  # type: ignore[misc] # breaking covariance
        """Wrap the object reference for this property."""
        obj = super().wrap_obj_ref(obj_ref)
        obj._attr_name = rich_text.snake_case(obj.name)
        return obj

    def delete(self) -> None:
        """Delete this property from the schema."""
        self._rename_prop(None)

    @property
    def id(self) -> str | None:
        """Return identifier of this property."""
        return raise_unset(self.obj_ref.id)

    @property
    def name(self) -> str:
        """Return name of this property."""
        if self._is_init and not is_unset(self.obj_ref.name):
            return self.obj_ref.name
        elif self._name is not None:
            return self._name
        else:
            msg = f'A name needs to be provided for `{self.__class__.__name__}`.'
            raise PropertyError(msg)

    @name.setter
    def name(self, new_name: str) -> None:
        """Set the name of this property."""
        self._rename_prop(new_name)

    @property
    def description(self) -> str | None:
        """Return the description of this property."""
        return self.obj_ref.description

    @property
    def attr_name(self) -> str:
        """Return the Python attribute name of the property in the schema."""
        if self._attr_name is None:
            msg = f'Property `{self.name}` has no assigned attribute name yet.'
            raise PropertyError(msg)
        return self._attr_name

    @attr_name.setter
    def attr_name(self, name: str) -> None:
        """Define the attribute name of the property within the schema."""
        self._attr_name = name

    @property
    def prop_value(self) -> type[PropertyValue]:
        """Return the corresponding property value of this property."""
        return PropertyValue._type_value_map[self.obj_ref.type]

    @property
    def readonly(self) -> bool:
        """Return if this property is read-only."""
        return self.prop_value.readonly


class Title(Property[obj_schema.Title], wraps=obj_schema.Title):
    """Defines the mandatory title property in a database."""


class Text(Property[obj_schema.RichText], wraps=obj_schema.RichText):
    """Defines a text property in a database."""


class Number(Property[obj_schema.Number], wraps=obj_schema.Number):
    """Defines a number property in a database."""

    def __init__(self, name: str | None = None, *, format: NumberFormat | str = NumberFormat.NUMBER):  # noqa: A002
        super().__init__(name, format=NumberFormat(format))

    @property
    def format(self) -> NumberFormat:
        """Return the number format of this number property."""
        return self.obj_ref.number.format

    @format.setter  # noqa: A003
    def format(self, new_format: NumberFormat | str) -> None:
        """Set the number format of this number property."""
        if isinstance(new_format, str):
            new_format = NumberFormat(new_format)
        self.obj_ref.number.format = new_format
        self._update_prop(self.obj_ref)


class Select(Property[obj_schema.Select], wraps=obj_schema.Select):
    """Defines a select property in a database."""

    def __init__(self, name: str | None = None, *, options: list[Option] | type[OptionNS]):
        if isinstance(options, type) and issubclass(options, OptionNS):
            options = options.to_list()

        option_objs = [option.obj_ref for option in options]
        super().__init__(name, options=option_objs)

    @property
    def options(self) -> list[Option]:
        """Return the options of this select property."""
        return [Option.wrap_obj_ref(option) for option in self.obj_ref.select.options]

    @options.setter
    def options(self, new_options: list[Option] | type[OptionNS]) -> None:
        """Set the options of this select property.

        !!! note
            Omitted options are removed from the property and new options
            will be added. Updating the `name` and `color` of an existing option
            is not supported via the API.
        """
        if isinstance(new_options, type) and issubclass(new_options, OptionNS):
            new_options = new_options.to_list()

        updates = check_for_updates(self.options, new_options)
        if updates:
            msg = f'Cannot update options of {self.name} property: {updates}'
            raise InvalidAPIUsageError(msg)

        self.obj_ref.select.options = [option.obj_ref for option in new_options]
        self._update_prop(self.obj_ref)


class MultiSelect(Property[obj_schema.MultiSelect], wraps=obj_schema.MultiSelect):
    """Defines a multi-select property in a database."""

    def __init__(self, name: str | None = None, *, options: list[Option] | type[OptionNS]):
        if isinstance(options, type) and issubclass(options, OptionNS):
            options = options.to_list()

        option_objs = [option.obj_ref for option in options]
        super().__init__(name, options=option_objs)

    @property
    def options(self) -> list[Option]:
        """Return the options of this multi-select property."""
        return [Option.wrap_obj_ref(option) for option in self.obj_ref.multi_select.options]

    @options.setter
    def options(self, new_options: list[Option] | type[OptionNS]) -> None:
        """Set the options of this multi-select property.

        !!! note
            Omitted options are removed from the property and new options
            will be added. Updating the `name` and `color` of an existing option
            is not supported via the API.
        """
        if isinstance(new_options, type) and issubclass(new_options, OptionNS):
            new_options = new_options.to_list()

        # Compare current and new options, handle changed attributes if needed
        diffs = check_for_updates(self.options, new_options)
        if diffs:
            msg = f'Cannot update options of {self.name} property: {diffs}'
            raise InvalidAPIUsageError(msg)

        self.obj_ref.multi_select.options = [option.obj_ref for option in new_options]
        self._update_prop(self.obj_ref)


class Status(Property[obj_schema.Status], wraps=obj_schema.Status):
    """Defines a status property in a database.

    The Notion API doesn't allow to create a property of this type.
    Sending it to the API with options and option groups defined results in an error
    about the existence of the keys `options` and `groups` and removing them
    creates a database with the property missing... ignorance is bliss.

    Also the Status configuration is not mentioned as a
    [Property Schema Object])https://developers.notion.com/reference/property-schema-object).

    It can still be used to check a schema.
    """

    allowed_at_creation = False

    def __init__(
        self,
        name: str | None = None,
        *,
        to_do: list[Option] | type[OptionNS] | None = None,
        in_progress: list[Option] | type[OptionNS] | None = None,
        complete: list[Option] | type[OptionNS] | None = None,
    ) -> None:
        def _normalize(group: list[Option] | type[OptionNS] | None) -> list[Option]:
            if group is None:
                return []
            elif isinstance(group, type) and issubclass(group, OptionNS):
                return group.to_list()
            else:
                return group

        to_do = _normalize(to_do)
        in_progress = _normalize(in_progress)
        complete = _normalize(complete)

        to_do_group = OptionGroup(OptionGroupType.TO_DO, options=to_do)
        in_progress_group = OptionGroup(OptionGroupType.IN_PROGRESS, options=in_progress)
        complete_group = OptionGroup(OptionGroupType.COMPLETE, options=complete)

        option_objs = [option.obj_ref for option in (*to_do, *in_progress, *complete)]
        group_objs = [group.obj_ref for group in (to_do_group, in_progress_group, complete_group)]
        super().__init__(name, options=option_objs, groups=group_objs)

    @property
    def options(self) -> list[Option]:
        """Return the options of this status property."""
        return [Option.wrap_obj_ref(option) for option in self.obj_ref.status.options]

    def _extract_groups(self) -> list[tuple[obj_schema.SelectGroup, list[Option]]]:
        """Extract options from a group."""

        def get_id(option: Option) -> str:
            try:
                return option.id
            except UnsetError:
                return option.name

        option_ids = {get_id(option): option for option in self.options}

        groups = []
        for group_obj in self.obj_ref.status.groups:
            # note that opt_id might be a name here, as explained in `OptionGroup``
            group_options = [option_ids[opt_id] for opt_id in set(group_obj.option_ids) & set(option_ids.keys())]
            groups.append((group_obj, group_options))

        return groups

    @property
    def groups(self) -> list[OptionGroup]:
        """Return the option groups of this status property."""
        return [OptionGroup.wrap_obj_ref(group, options=options) for group, options in self._extract_groups()]


class Date(Property[obj_schema.Date], wraps=obj_schema.Date):
    """Defines a date property in a database."""


class Person(Property[obj_schema.People], wraps=obj_schema.People):
    """Defines a person/people property in a database."""


class Files(Property[obj_schema.Files], wraps=obj_schema.Files):
    """Defines a files property in a database."""


class Checkbox(Property[obj_schema.Checkbox], wraps=obj_schema.Checkbox):
    """Defines a checkbox property in database."""


class Email(Property[obj_schema.Email], wraps=obj_schema.Email):
    """Defines an e-mail property in a database."""


class URL(Property[obj_schema.URL], wraps=obj_schema.URL):
    """Defines a URL property in a database."""


class Phone(Property[obj_schema.PhoneNumber], wraps=obj_schema.PhoneNumber):
    """Defines a phone number property in a database."""


class Formula(Property[obj_schema.Formula], wraps=obj_schema.Formula):
    """Defines a formula property in a database.

    Currently the formula expression cannot reference other formula properties, e.g. `prop("other formula")`
    This is a limitation of the Notion API.
    """

    @property
    def formula(self) -> str:
        """Return the formula of this property."""
        return self.obj_ref.formula.expression

    @formula.setter
    def formula(self, formula: str) -> None:
        """Set the formula of this property."""
        self.obj_ref.formula.expression = formula
        self._update_prop(self.obj_ref)


class Relation(Property[obj_schema.Relation], wraps=obj_schema.Relation):
    """Relation to another database."""

    _rel_schema: type[Schema] | None = None  # other schema, i.e. of the target database
    _two_way_prop: Property | str | None = None  # other property, i.e. of the target database

    def __init__(
        self, name: str | None = None, *, schema: type[Schema] | None = None, two_way_prop: Relation | str | None = None
    ):
        self._name = name
        self._owner = None
        # Note that we don't call super().__init__ here since we only know how to build the obj_ref later.
        if two_way_prop and not schema:
            msg = '`schema` needs to be provided if `two_way_prop` is set!'
            raise RuntimeError(msg)

        if isinstance(schema, Property):
            msg = 'Please provide a schema, not a property! Use `two_way_prop` to specify a property.'
            raise ValueError(msg)

        self._rel_schema = schema

        if two_way_prop is not None:
            if not isinstance(two_way_prop, Relation | str):
                msg = 'The two-way property parameter needs to be of type Relation or str!'
                raise ValueError(msg)

            if isinstance(two_way_prop, Relation) and two_way_prop._rel_schema is not None:
                msg = f'The two-way property {two_way_prop.name} must not reference a schema itself!'
                raise ValueError(msg)

            self._two_way_prop = two_way_prop

    @property
    def _is_init_ready(self) -> bool:
        """Determines if the Relation can be initialized"""
        return not (self._rel_schema is SelfRef or self._is_two_way_target)

    def _make_obj_ref(self) -> obj_schema.Relation:
        """Create the low-level object reference for this relation."""
        try:
            db = self.schema.get_db()
        except SchemaNotBoundError as e:
            msg = f"A database with schema '{self.schema.__name__}' needs to be created first!"
            raise RelationError(msg) from e

        if self._two_way_prop:
            obj_ref = obj_schema.DualPropertyRelation.build_relation(db.id)
        else:
            obj_ref = obj_schema.SinglePropertyRelation.build_relation(db.id)
        return obj_ref

    @property
    def obj_ref(self) -> obj_schema.Relation:
        """Initialize the low-level object references for this relation."""
        if not self._is_init:
            # Delayed construction of obj_ref to assure that a self-reference is resolved.
            self._obj_ref = self._make_obj_ref()
        return self._obj_ref

    @obj_ref.setter
    def obj_ref(self, new_obj_ref: obj_schema.Relation) -> None:
        self._obj_ref = new_obj_ref

    @property
    def schema(self) -> type[Schema]:
        """Schema of the relation database."""
        if self._rel_schema is None and self._get_owner().is_bound():
            session = get_active_session()
            self._rel_schema = session.get_db(self.obj_ref.relation.database_id).schema
        elif self._rel_schema is not None and self._rel_schema is not SelfRef:
            return self._rel_schema
        else:
            msg = 'The relation is not yet related to another schema!'
            raise RelationError(msg)
        return self._rel_schema

    @schema.setter
    def schema(self, new_schema: type[Schema]) -> None:
        """Set the schema of the relation database."""
        if self.is_two_way:
            new_rel = obj_schema.DualPropertyRelation.build_relation(new_schema.get_db().id)
        else:
            new_rel = obj_schema.SinglePropertyRelation.build_relation(new_schema.get_db().id)
        self.obj_ref.relation = new_rel.relation
        self._update_prop(self.obj_ref)
        self._rel_schema = new_schema

    def _rename_two_way_prop(self, new_prop_name: str) -> None:
        """Rename the two-way property in the target schema.

        This is necessary as a two-way relation is created with a default name,
        which is rather a bug in the Notion API itself.
        """
        two_way_prop_rel = cast(obj_schema.DualPropertyRelation, self.obj_ref.relation)
        tmp_prop_name = raise_unset(two_way_prop_rel.dual_property.synced_property_name)
        db = self.schema.get_db()
        db.reload(rebind_schema=False)
        if (back_ref_prop_type := db.schema[tmp_prop_name]) is not None:
            back_ref_prop_type.name = new_prop_name  # rename the property in the target schema
            back_ref_prop_type._attr_name = rich_text.snake_case(new_prop_name)

    @property
    def two_way_prop(self) -> Property | None:
        """Return the target property object of a two-way relation."""
        if self._two_way_prop:
            if isinstance(self._two_way_prop, str):
                self._two_way_prop = self._get_owner().get_prop(self._two_way_prop)
            return self._two_way_prop
        elif self._is_init and isinstance(self.obj_ref.relation, obj_schema.DualPropertyRelation):
            prop_name = raise_unset(self.obj_ref.relation.dual_property.synced_property_name)
            return self.schema.get_prop(prop_name) if prop_name else None
        else:
            return None

    @two_way_prop.setter
    def two_way_prop(self, prop_name: str | None) -> None:
        """Set the target property as string of a two-way relation.

        The `new_prop_name` is the name of the property in the target schema.
        """

        if (db := self.schema.get_db()) is None:
            msg = 'The target schema of the relation is not bound to a database'
            raise RelationError(msg)

        if prop_name is None:  # delete the two-way property
            if self.two_way_prop is None:
                return
            target_schema = self.schema
            target_two_way_prop = self.two_way_prop.name
            new_rel = obj_schema.SinglePropertyRelation.build_relation(db.id)
            self.obj_ref.relation = new_rel.relation
            self.obj_ref = self._update_prop(self.obj_ref)
            # Strangely enough, the two-way property is not removed from the target schema
            # also it is no longer a two-way relation.
            del target_schema[target_two_way_prop]
        else:
            new_rel = obj_schema.DualPropertyRelation.build_relation(db.id)
            self.obj_ref.relation = cast(obj_schema.DualPropertyRelation, new_rel.relation)
            self.obj_ref = self._update_prop(self.obj_ref)
            self._rename_two_way_prop(prop_name)

        self._rel_schema = None  # Don't rely on the initialization of the schema

    @property
    def is_two_way(self) -> bool:
        """Determine if this relation is a two-way relation."""
        return self._two_way_prop is not None or (
            self._is_init and isinstance(self.obj_ref.relation, obj_schema.DualPropertyRelation)
        )

    @property
    def is_self_ref(self) -> bool:
        """Determines if this relation is self referencing the same schema."""
        return (self._rel_schema is SelfRef) or (self._rel_schema is self._get_owner())

    @property
    def _is_two_way_target(self) -> bool:
        """Determines if this relation is a target of a two-way relation."""
        return self._rel_schema is None

    def _update_bwd_rel(self) -> None:
        """Change the default name of a two-way relation target to the defined one."""
        if not isinstance(self.obj_ref.relation, obj_schema.DualPropertyRelation):
            msg = f'Trying to inialize backward relation for one-way relation {self.name}'
            raise SchemaError(msg)

        obj_synced_property_name = self.obj_ref.relation.dual_property.synced_property_name
        two_way_prop_name = self.two_way_prop.name if self.two_way_prop else None

        if obj_synced_property_name != two_way_prop_name:
            session = get_active_session()

            # change the old default name in the target schema to what was passed during initialization
            other_db = self.schema.get_db()

            if is_unset(prop_id := self.obj_ref.relation.dual_property.synced_property_id):
                msg = 'No synced property ID found in the relation object'
                raise SchemaError(msg)

            schema_dct = {prop_id: obj_schema.RenameProp(name=two_way_prop_name)}
            session.api.databases.update(db=other_db.obj_ref, schema=schema_dct)
            other_db.schema._set_obj_refs()

            our_db = self._get_owner().get_db()
            session.api.databases.update(db=our_db.obj_ref, schema={})  # sync obj_ref
            our_db.schema._set_obj_refs()


class Rollup(Property[obj_schema.Rollup], wraps=obj_schema.Rollup):
    """Defines the rollup property in a database.

    If the relation propery is a self-referencing relation, i.e. `uno.PropType.Relation(uno.SelfRef)` in the schema,
    then the `property` must be a `str` of the corresponding property name.
    """

    def __init__(
        self,
        name: str | None = None,
        *,
        relation: Relation,
        rollup: Property,
        calculate: AggFunc | str = AggFunc.SHOW_ORIGINAL,
    ):
        calculate = AggFunc.from_alias(calculate) if not isinstance(calculate, AggFunc) else calculate

        # ToDo: One could check here if property really is a property in the database where relation points to
        super().__init__(name, relation=relation.name, property=rollup.name, function=calculate)

    @property
    def _is_init_ready(self) -> bool:
        """Determines if the relation of the rollup is ready to be initialized."""
        try:
            return self.relation_prop._is_init_ready
        except SchemaError:  # DB is not bound yet
            return False

    @property
    def is_self_ref(self) -> bool:
        """Determines if this rollup is self-referencing the same schema."""
        return self.relation_prop.is_self_ref

    @property
    def relation_prop(self) -> Relation:
        """Return the relation property object of the rollup."""
        rel_prop = self._get_owner().get_prop(self.obj_ref.rollup.relation_property_name)
        if isinstance(rel_prop, Relation):
            return rel_prop
        else:
            msg = f'Rollup property {self.name} is not bound to a Relation property'
            raise RollupError(msg)

    @property
    def rollup_prop(self) -> Property:
        """Return the rollup property object of the rollup."""
        return self._get_owner().get_prop(self.obj_ref.rollup.rollup_property_name)

    @property
    def calculate(self) -> AggFunc:
        """Return the aggregation function of the rollup."""
        return self.obj_ref.rollup.function


class CreatedTime(Property[obj_schema.CreatedTime], wraps=obj_schema.CreatedTime):
    """Defines the created-time property in a database."""


class CreatedBy(Property[obj_schema.CreatedBy], wraps=obj_schema.CreatedBy):
    """Defines the created-by property in a database."""


class LastEditedBy(Property[obj_schema.LastEditedBy], wraps=obj_schema.LastEditedBy):
    """Defines the last-edited-by property in a database."""


class LastEditedTime(Property[obj_schema.LastEditedTime], wraps=obj_schema.LastEditedTime):
    """Defines the last-edited-time property in a database."""


class ID(Property[obj_schema.UniqueID], wraps=obj_schema.UniqueID):
    """Defines a unique ID property in a database.

    !!! note
        The prefix of the ID will be capitalized by Notion automatically.
        If no prefix is provided, Notion will assign one automatically.
        If a prefix is provided, the prefix itself must be unique in the workspace.
    """

    def __init__(self, name: str | None = None, *, prefix: str | None = None):
        super().__init__(name=name, prefix=self._norm_prefix(prefix))

    @staticmethod
    def _norm_prefix(prefix: str | None) -> str | None:
        """Normalize the prefix to be alphanumeric and uppercase."""
        if prefix is None or prefix == '':
            return None
        elif not prefix.isalnum():
            msg = 'The prefix of a unique ID must be alphanumeric!'
            raise ValueError(msg)
        else:
            return prefix.upper()

    @property
    def prefix(self) -> str:
        """Return the prefix of the unique ID."""
        opt_prefix = self.obj_ref.unique_id.prefix
        return '' if opt_prefix is None else opt_prefix

    @prefix.setter
    def prefix(self, prefix: str | None) -> None:
        """Set the prefix of the unique ID."""
        self.obj_ref.unique_id.prefix = self._norm_prefix(prefix)
        self._update_prop(self.obj_ref)


class Verification(Property[obj_schema.Verification], wraps=obj_schema.Verification):
    """Defines a unique ID property in a database."""

    allowed_at_creation = False


class Button(Property[obj_schema.Button], wraps=obj_schema.Button):
    """Defines a button property in a database."""

    allowed_at_creation = False


class PropType:
    """Namespace class of all property types of a database for easier access."""

    Title = Title
    Text = Text
    Number = Number
    Select = Select
    MultiSelect = MultiSelect
    Status = Status
    Date = Date
    Person = Person
    Files = Files
    Checkbox = Checkbox
    Email = Email
    URL = URL
    Phone = Phone
    Formula = Formula
    Relation = Relation
    Rollup = Rollup
    CreatedTime = CreatedTime
    CreatedBy = CreatedBy
    LastEditedTime = LastEditedTime
    LastEditedBy = LastEditedBy
    ID = ID
    Verification = Verification


class SchemaType(ABCMeta):
    """Metaclass for the schema of a database.

    This makes the schema class itself more user-friendly by providing a `__magic__` methods, e.g.
    letting it behave like a dictionary for the properties although it is a class, not an instance.
    """

    # ToDo: When mypy is smart enough to understand metaclasses, we can remove the `type: ignore` comments
    def __new__(metacls, name: str, bases: tuple[type, ...], namespace: dict[str, object], **kwargs: Any) -> SchemaType:
        # This collects all schema properties under `_props`. Using __prepare__ doesn't work as referencing
        # the class itself would not be possible at that stage any more as __getattr__ is not used at that point.
        # Thus, we create the class normally and change it only afterwards when __prepare__ is already done.
        props = cast(list[Property], namespace.setdefault('_props', []))

        for b in bases:
            for prop in getattr(b, '_props', []):
                prop = cast(Property, deepcopy(prop))
                props.append(prop)

        for attr, val in list(namespace.items()):
            if isinstance(val, Property):
                del namespace[attr]
                val._attr_name = attr
                props.append(val)

        cls = super().__new__(metacls, name, bases, dict(namespace), **kwargs)
        for prop in props:
            prop._owner = cast(type[Schema], cls)

        return cls

    def __str__(cls: type[Schema]) -> str:  # type: ignore[misc]
        # We can only overwrite __str__ for a class in a metaclass
        return cls.as_table(tablefmt='simple')

    def __getitem__(cls: type[Schema], prop_name: str) -> Property:  # type: ignore[misc]
        return cls.get_prop(prop_name)

    def __delitem__(cls: type[Schema], prop_name: str) -> None:  # type: ignore[misc]
        cls.get_prop(prop_name).delete()

    def __setitem__(cls: type[Schema], prop_name: str, prop_type: Property) -> None:  # type: ignore[misc]
        if prop_type._name is not None:
            msg = (
                f'Property `{prop_name}` already has a name and thus the `name` parameter of `{prop_type}` must '
                'not be set or use `prop.name = "..."` to change the name of a property.'
            )
            raise PropertyError(msg)

        # handle updating existing properties
        if (curr_prop := cls.get_prop(prop_name, default=None)) is not None:
            if prop_type._attr_name is None:
                prop_type._attr_name = curr_prop._attr_name
            curr_prop.delete()

        prop_type._name = prop_name
        prop_type._owner = cls
        if prop_type._attr_name is None:
            prop_type._attr_name = rich_text.snake_case(prop_name)

        session = get_active_session()
        session.api.databases.update(db=cls.get_db().obj_ref, schema={prop_name: prop_type.obj_ref})

        cls._props.append(prop_type)
        cls._set_obj_refs()

        if isinstance(prop_type, Relation) and (prop_type._two_way_prop is not None):
            prop_type._rename_two_way_prop(prop if isinstance(prop := prop_type._two_way_prop, str) else prop.name)

    def __getattr__(cls: type[Schema], name: str) -> Property:  # type: ignore[misc]
        attr_name_props = SList([prop for prop in cls.get_props() if prop._attr_name == name])
        try:
            return attr_name_props.item()
        except MultipleItemsError as e:
            non_unique_attrs = ', '.join(str(attr) for attr in attr_name_props)
            msg = (
                f'Attribute {name} is not unique for properties {non_unique_attrs}! '
                'Use indexing, i.e. [...], to access the property.'
            )
            raise AttributeError(msg) from e
        except EmptyListError as e:
            msg = f'Schema has no property with attribute name {name}.'
            raise AttributeError(msg) from e

    def __setattr__(cls: type[Schema], name: str, value: Any) -> None:  # type: ignore[misc]
        # Overwrite runtime setting of class attributes
        if isinstance(value, Property):
            curr_attr = getattr(cls, name, None)
            if curr_attr is None:  # adding a new property
                prop_name = value.name
                if prop_name is None:
                    msg = f'Property `{name}` must have a name.'
                    raise PropertyError(msg)
                else:
                    value._name = None  # __setitem__ needs _name to be unset
                    value._attr_name = name
                    cls[prop_name] = value
            elif isinstance(curr_attr, Property):  # updating an existing property
                cls[curr_attr.name] = value
            else:
                msg = f'Cannot override non-property `{name}` of type `{type(curr_attr)}`.'
                raise PropertyError(msg)
        else:
            super().__setattr__(name, value)  # type: ignore[misc]  # no clue why this results in a type problem

    def __len__(cls: type[Schema]) -> int:  # type: ignore[misc]
        return len(cls.get_props())

    def __iter__(cls: type[Schema]) -> Iterator[Property]:  # type: ignore[misc]
        return (prop for prop in cls.get_props())


class SchemaModel(BaseModel):
    """Base Pydantic model for schemas to validate pages within a database."""

    model_config = ConfigDict(arbitrary_types_allowed=True, extra='forbid')
    _db_title: str | None = PrivateAttr(default=None)
    _db_id: str | None = PrivateAttr(default=None)

    def to_dict(self) -> dict[str, Any]:
        """Convert the Pydantic model to a dictionary."""
        result = {}
        for name, field in self.__class__.model_fields.items():
            value = getattr(self, name)
            name = field.alias or name
            if isinstance(value, PropertyValue):
                result[name] = value.obj_ref
        return result

    def __repr__(self) -> str:
        base = super().__repr__()
        attrs = base[len(self.__class__.__name__) :]
        title = 'Untitled' if self._db_title is None else f' ({self._db_title})'
        return f'{self.__class__.__name__}[{title}]{attrs}'


class Schema(metaclass=SchemaType):
    """Base class for the schema of a database."""

    _db_title: rich_text.Text | None
    _db_id: str | None
    _db_desc: rich_text.Text | None
    _database: Database | None = None
    _props: list[Property]

    def __init_subclass__(cls, db_title: str | None = None, db_id: str | None = None, **kwargs: Any):
        if db_title is not None:
            db_title = rich_text.Text(db_title)
        cls._db_title = db_title
        cls._db_id = db_id
        cls._db_desc = rich_text.Text(cls.__doc__) if cls.__doc__ is not None else None
        cls._validate()
        super().__init_subclass__(**kwargs)

    @classmethod
    def _validate(cls) -> None:
        """Validate the schema before creating a database."""
        try:
            cls.get_title_prop()
        except EmptyListError as e:
            msg = 'A property of type `Title` is mandatory'
            raise SchemaError(msg) from e
        except MultipleItemsError as e:
            msg = 'Only one property of type `Title` is allowed'
            raise SchemaError(msg) from e

        name_counter = Counter(prop.name for prop in cls.get_props())
        if duplicates := [prop for prop, count in name_counter.items() if count > 1]:
            msg = f'Properties {", ".join(duplicates)} are defined more than once'
            raise SchemaError(msg)

        prop_counter = Counter(prop.__class__.__name__ for prop in cls.get_props())
        if prop_counter.get(ID.__name__, 0) > 1:
            msg = 'Only one property of type `ID` is allowed'
            raise SchemaError(msg)

    @classmethod
    def create(cls, **kwargs: Any) -> Page:
        """Create a page using this schema with a bound database."""
        return cls.get_db().create_page(**kwargs)

    @classmethod
    def get_props(cls) -> list[Property]:
        """Get all properties of this schema."""
        return cls._props

    @overload
    @classmethod
    def get_prop(cls, prop_name: str, *, default: UnsetType = ...) -> Property: ...
    @overload
    @classmethod
    def get_prop(cls, prop_name: str, *, default: T) -> Property | T: ...

    @classmethod
    def get_prop(cls, prop_name: str, *, default: object = Unset) -> Property | T:
        """Get a specific property from this schema assuming that property names are unique."""
        try:
            return SList([prop for prop in cls.get_props() if prop.name == prop_name]).item()
        except EmptyListError as e:
            if default is Unset:
                if cls.is_bound():
                    msg = f'Property `{prop_name}` not found in schema of `{cls._database}`.'
                else:
                    msg = f'Property `{prop_name}` not found in unbound schema.'
                raise SchemaError(msg) from e
            return cast(T, default)

    @classmethod
    def has_prop(cls, prop_name: str) -> bool:
        """Check if a property exists in this schema."""
        return cls.get_prop(prop_name, default=None) is not None

    @classmethod
    def get_ro_props(cls) -> list[Property]:
        """Get all read-only properties of this schema."""
        return [prop for prop in cls.get_props() if prop.readonly]

    @classmethod
    def get_rw_props(cls) -> list[Property]:
        """Get all writeable properties of this schema."""
        return [prop for prop in cls.get_props() if not prop.readonly]

    @classmethod
    def to_pydantic_model(cls, *, with_ro_props: bool = False) -> type[SchemaModel]:
        """Return a Pydantic model of this schema for validation.

        This is useful for instance when writing a web API that receives data that should be validated
        before it is passed to Ultimate Notion. The actual values are converted to `PropertyValue`
        and thus `value` needs to be called to retrieve the actual Python type.

        If `with_ro_props` is set to `True`, read-only properties are included in the model.
        """

        # ToDo: Validate the categories in Select and MultiSelect using pydantic!
        def pytype_to_prop_value(py_type: Any, *, prop_value: type[PropertyValue]) -> PropertyValue:
            """Convert a Python type to a PropertyValue."""
            return py_type if isinstance(py_type, PropertyValue) else prop_value(py_type)

        kwargs: dict[str, Any] = {
            prop.attr_name: Annotated[prop.prop_value, Field(default=None, alias=prop.name)]
            for prop in cls.get_rw_props()
        }
        validators: dict[str, Any] = {
            f'{prop.attr_name}_validator': field_validator(prop.attr_name, mode='before')(
                partial(pytype_to_prop_value, prop_value=prop.prop_value)
            )
            for prop in cls.get_rw_props()
        }

        if with_ro_props:
            kwargs |= {
                prop.attr_name: Annotated[prop.prop_value, Field(alias=prop.name)] for prop in cls.get_ro_props()
            }
            validators |= {
                f'{prop.attr_name}_validator': field_validator(prop.attr_name, mode='before')(
                    partial(pytype_to_prop_value, prop_value=prop.prop_value)
                )
                for prop in cls.get_ro_props()
            }

        model = create_model(
            SchemaModel.__name__,
            __validators__=validators,
            __base__=SchemaModel,
            **kwargs,
        )
        model._db_title = cls._db_title
        model._db_id = cls._db_id
        return model

    @classmethod
    def to_dict(cls) -> dict[str, Property]:
        """Convert this schema to a dictionary of property names and corresponding types."""
        return {prop.name: prop for prop in cls.get_props()}

    @classmethod
    def as_table(cls, tablefmt: str | None = None) -> str:
        """Return the schema in a given string table format.

        Some table formats:

        - plain: no pseudographics
        - simple: Pandoc's simple table, i.e. only dashes to separate header from content
        - github: GitHub flavored Markdown
        - simple_grid: uses dashes & pipes to separate cells
        - html: standard html markup

        Find more table formats under: https://github.com/astanin/python-tabulate#table-format
        """
        if tablefmt is None:
            tablefmt = 'html' if is_notebook() else 'simple'

        headers = ['Name', 'Property', 'Attribute']
        rows = []
        for prop in cls.get_props():
            rows.append((prop.name, prop, prop.attr_name))

        return tabulate(rows, headers=headers, tablefmt=tablefmt)

    @classmethod
    def show(cls, *, simple: bool | None = None) -> None:
        """Show the schema as html or as simple table."""
        if simple:
            tablefmt = 'simple'
        elif simple is None:
            tablefmt = 'html' if is_notebook() else 'simple'
        else:
            tablefmt = 'html'

        table_str = cls.as_table(tablefmt)

        if is_notebook() and (tablefmt == 'html'):
            from IPython.display import display_html  # noqa: PLC0415

            display_html(table_str)
        else:
            print(table_str)  # noqa: T201

    @classmethod
    def _repr_html_(cls) -> str:  # noqa: PLW3201
        """Called by JupyterLab automatically to display this schema."""
        return cls.as_table(tablefmt='html')

    @classmethod
    def get_title_prop(cls) -> Property:
        """Returns the property holding the title of the pages."""
        return SList(prop for prop in cls.get_props() if isinstance(prop, Title)).item()

    @classmethod
    def assert_consistency_with(cls, other_schema: type[Schema], *, during_init: bool = False) -> None:
        """Assert that this schema is consistent with another schema."""
        own_schema_dct = cls.to_dict()
        other_schema_dct = other_schema.to_dict()

        if during_init:
            # backward relation was not yet initialised in the other schema (during the creation of the data model)
            # or self-referencing relation was not yet initialised
            own_schema_dct = {
                name: prop_type
                for name, prop_type in own_schema_dct.items()
                if not (
                    isinstance(prop_type, Relation) and (prop_type.is_two_way)  # synced_prop_name default name
                )
            }

            other_schema_dct = {
                name: prop_type
                for name, prop_type in other_schema_dct.items()
                if not (
                    (isinstance(prop_type, Relation) and (prop_type.is_two_way))
                    or (isinstance(prop_type, Relation) and (prop_type._is_two_way_target or prop_type.is_self_ref))
                    or (isinstance(prop_type, Rollup) and prop_type.is_self_ref)
                )
            }

        if other_schema_dct != own_schema_dct:
            props_added, props_removed, props_changed = dict_diff_str(own_schema_dct, other_schema_dct)
            msg = 'Provided schema is not consistent with the current schema of the database:\n'
            if props_added:
                msg += f'- added: {", ".join(props_added)}\n'
            if props_removed:
                msg += f'- removed: {", ".join(props_removed)}\n'
            if props_changed:
                msg += f'- changed: {", ".join(props_changed)}\n'
            raise SchemaError(msg)

    @classmethod
    def get_db(cls) -> Database:
        """Get the database that is bound to this schema."""
        if cls._database is not None:  # is_bound() cannot be used here due to type checker
            return cls._database
        else:
            raise SchemaNotBoundError(cls)

    @classmethod
    def _bind_db(cls, db: Database) -> None:
        """Bind this schema to the corresponding database for back-reference without setting it in `db`."""
        # Needed to break the recursion when setting a schema in Database
        cls._database = db
        cls._set_obj_refs()

    @classmethod
    def bind_db(cls, db: Database | None = None) -> None:
        """Bind this schema to the corresponding database for back-reference and vice versa.

        If `None` (default) is passed, search for the database using `db_id` or `db_title`
        and bind it to this schema.
        """
        if db is None:
            sess = get_active_session()
            if cls._db_id is not None:
                db = sess.get_db(cls._db_id)
            elif cls._db_title is not None:
                db = sess.search_db(cls._db_title).item()
            else:
                msg = 'Neither a database ID nor a title is set to bind the schema automatically.'
                raise InvalidAPIUsageError(msg)

        db.schema = cls
        cls._bind_db(db)

    @classmethod
    def is_bound(cls) -> bool:
        """Determines if the schema is bound to a database."""
        return cls._database is not None

    @classmethod
    def _get_fwd_rels(cls) -> list[Relation]:
        return [
            prop
            for prop in cls.get_props()
            if isinstance(prop, Relation) and not (prop._is_two_way_target or prop.is_self_ref)
        ]

    @classmethod
    def _get_self_refs(cls) -> list[Relation]:
        """Get all self-referencing relation properties."""
        return [prop for prop in cls.get_props() if isinstance(prop, Relation) and prop.is_self_ref]

    @classmethod
    def _has_self_refs(cls) -> bool:
        """Determine if self-referencing relation properties are present."""
        return bool([prop for prop in cls.get_props() if isinstance(prop, Relation) and prop.is_self_ref])

    @classmethod
    def _init_self_refs(cls) -> None:
        """Initialise all forward self-referencing relations."""
        if not cls._has_self_refs():
            return
        db = cls.get_db()  # raises if not bound!
        for prop_type in cls._get_self_refs():
            prop_type._rel_schema = cls  # replace placeholder `SelfRef` with this schema

        new_props_schema = {prop.name: prop.obj_ref for prop in cls._get_self_refs()} or None
        session = get_active_session()
        session.api.databases.update(db.obj_ref, schema=new_props_schema)
        cls._set_obj_refs()

    @classmethod
    def _init_self_ref_rollups(cls) -> None:
        """Initialise all rollup properties that reference the same schema."""
        if not cls._has_self_refs():
            return
        db = cls.get_db()  # raises if not bound!

        self_ref_rollups = [prop for prop in cls.get_props() if isinstance(prop, Rollup) and prop.is_self_ref]
        new_props_schema = {prop.name: prop.obj_ref for prop in self_ref_rollups} or None
        session = get_active_session()
        session.api.databases.update(db.obj_ref, schema=new_props_schema)
        cls._set_obj_refs()

    @classmethod
    def _update_bwd_rels(cls) -> None:
        """Update the default property name in case of a two-way relation in the external target schema.

        By default the property in the target schema is named "Related to <this_database> (<this_field>)"
        which is then set to the name specified as backward relation.
        """
        for prop_type in cls.to_dict().values():
            if isinstance(prop_type, Relation) and prop_type.is_two_way:
                prop_type._update_bwd_rel()

    @classmethod
    def _set_obj_refs(cls) -> None:
        """Map obj_refs from the properties of the schema to obj_ref.properties of the bound database."""
        db_props_dct = cls.get_db().obj_ref.properties
        for prop_name, prop_type in cls.to_dict().items():
            obj_ref = db_props_dct.get(prop_name)
            if obj_ref:
                prop_type.obj_ref = obj_ref


class DefaultSchema(Schema):
    # Default database schema of Notion.
    #
    # As inferred by just creating an empty database in the Notion UI.
    # NOTE: Use no docstring here, otherwise it will be used as the database description.

    name = Title('Name')


class SelfRef(Schema):
    """Target schema for self-referencing database relations."""

    _ = Title('title')  # mandatory title property, used for nothing.



"""Google Tasks API adapters."""

from ultimate_notion.adapters.google.tasks.client import GTasksClient
from ultimate_notion.adapters.google.tasks.sync import SyncGTasks

__all__ = ['GTasksClient', 'SyncGTasks']



"""Syncs a Notion database with a Google Tasks task list."""
# ruff: noqa: PLR6301

from __future__ import annotations

from datetime import date, datetime
from typing import Any

from ultimate_notion import Database, Page, Property
from ultimate_notion.adapters.google.tasks.client import GTask, GTaskList
from ultimate_notion.adapters.sync import ID, ConflictMode, SyncTask
from ultimate_notion.core import get_active_session
from ultimate_notion.utils import str_hash


class SyncGTasks(SyncTask):
    """Syncs a Notion database with a Google Tasks task list."""

    def __init__(
        self,
        *,
        notion_db: Database,
        tasklist: GTaskList,
        completed_col: Property | str,
        completed_val: Any,
        not_completed_val: Any,
        due_col: Property | str,
        name: str = 'SyncGTasks',
        conflict_mode: ConflictMode = ConflictMode.NEWER,
    ):
        if isinstance(completed_col, Property):
            completed_col = completed_col.name
        if isinstance(due_col, Property):
            due_col = due_col.name

        self.notion_db = notion_db
        self.tasklist = tasklist
        self.completed_col = completed_col
        self.completed_val = completed_val
        self.not_completed_val = not_completed_val
        self.due_col = due_col
        self.title_col = self.notion_db.schema.get_title_prop().name

        attr_map = {
            self.title_col: 'title',
            self.completed_col: 'is_completed',
            self.due_col: 'due',
        }
        super().__init__(name=name, attr_map=attr_map, conflict_mode=conflict_mode)

    def get_notion_objects(self) -> list[Page]:
        """Get all pages from database."""
        # We remove all cached pages from the database to make sure we get the latest version
        cache = get_active_session().cache
        cached_pages = [
            obj_id for obj_id, obj in cache.items() if isinstance(obj, Page) and obj.parent_db == self.notion_db
        ]
        for page_id in cached_pages:
            del cache[page_id]
        return self.notion_db.get_all_pages().to_pages()

    def get_other_objects(self) -> list[GTask]:
        """Get all Google Taks from Tasklist."""
        return self.tasklist.all_tasks()

    def notion_timestamp(self, obj: Page) -> datetime:
        """Get the timestamp of the Notion page."""
        return obj.last_edited_time

    def other_timestamp(self, obj: GTask) -> datetime:
        """Get the timestamp of the Google Task."""
        return obj.updated

    def notion_id(self, obj: Page) -> ID:
        """Get the ID of the Notion page."""
        return str(obj.id)

    def other_id(self, obj: GTask) -> ID:
        """Get the ID of the Google Task."""
        return obj.id

    def notion_hash(self, obj: Page) -> str:
        """Get the hash of the Notion page for object mapping/linking."""
        return str_hash(obj.title or '')

    def other_hash(self, obj: GTask) -> str:
        """Get the hash of the other object for object mapping/linking."""
        return str_hash(obj.title or '')

    def notion_to_dict(self, obj: Page) -> dict[str, Any]:
        """Convert a Notion object to a dictionary."""
        due_date: datetime | date | None = obj.props[self.due_col]
        if isinstance(due_date, datetime):
            due_date = due_date.date()
        return {
            self.title_col: obj.title,
            self.completed_col: obj.props[self.completed_col] == self.completed_val,
            self.due_col: due_date,
        }

    def other_to_dict(self, obj: GTask) -> dict[str, Any]:
        """Convert another object to a dictionary."""
        return {
            'title': obj.title,
            'is_completed': obj.is_completed,
            'due': obj.due if obj.due is None else obj.due.date(),
        }

    def notion_update_obj(self, obj: Page, attr: str, value: Any) -> None:
        """Set an attribute of the Notion object, e.g. page."""
        if attr == self.completed_col:
            if value:
                obj.props[attr] = self.completed_val
            else:
                obj.props[attr] = self.not_completed_val
        else:
            obj.props[attr] = value

    def other_update_obj(self, obj: GTask, attr: str, value: Any) -> None:
        """Set an attribute of the other object."""
        setattr(obj, attr, value)

    def notion_delete_obj(self, obj: Page) -> None:
        """Delete the page."""
        obj.delete()

    def other_delete_obj(self, obj: GTask) -> None:
        """Delete the other object."""
        obj.delete()

    def notion_create_obj(self, **kwargs: Any) -> Page:
        """Create a new page."""
        kwargs[self.completed_col] = self.completed_val if kwargs[self.completed_col] else self.not_completed_val
        attr_kwargs = {
            self.notion_db.schema.get_prop(key).attr_name: value for key, value in kwargs.items() if value is not None
        }
        page = self.notion_db.create_page(**attr_kwargs)
        return page

    def other_create_obj(self, **kwargs: Any) -> GTask:
        """Create a new other object."""
        task = self.tasklist.create_task(title=kwargs['title'], due=kwargs['due'])
        task.is_completed = kwargs['is_completed']
        return task



"""Dealing with emoji objects of the Notion API."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, TypeVar
from uuid import UUID

from emoji import demojize, emojize, is_emoji

from ultimate_notion.core import Wrapper, get_repr
from ultimate_notion.errors import InvalidAPIUsageError
from ultimate_notion.obj_api import objects as objs

TO = TypeVar('TO', bound=objs.TypedObject)


class EmojiBase(Wrapper[TO], str, ABC, wraps=objs.TypedObject):
    """Base class for emoji objects, which behave like str."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Return the name of the emoji."""
        raise NotImplementedError

    @property
    def to_code(self) -> str:
        """Represent the emoji as :shortcode:, e.g. :smile:"""
        return f':{self.name}:'

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, EmojiBase | str):
            return str(self) == str(other)
        else:
            return NotImplemented

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by JupyterLab automatically to display this file."""
        return str(self)


class Emoji(EmojiBase[objs.EmojiObject], wraps=objs.EmojiObject):
    """Unicode emoji object which behaves like str."""

    def __init__(self, emoji: str) -> None:
        if not is_emoji(emoji):
            emoji = emojize(emoji)
        if not is_emoji(emoji):
            msg = f'Invalid emoji string: {emoji}'
            raise ValueError(msg)
        self.obj_ref = objs.EmojiObject.build(emoji)

    @property
    def name(self) -> str:
        """Return the name of the emoji."""
        return demojize(self.obj_ref.emoji).strip(':')

    def __repr__(self) -> str:
        # we return the unicode emoji to the user for convenience in REPL
        return self.obj_ref.emoji

    def __str__(self) -> str:
        return self.obj_ref.emoji


class CustomEmoji(EmojiBase[objs.CustomEmojiObject], wraps=objs.CustomEmojiObject):
    """Custom emoji object which behaves like str."""

    def __init__(self) -> None:
        msg = 'To create a new custom emoji, go to `Settings` -> `Emoji` in Notion.'
        raise InvalidAPIUsageError(msg)

    @property
    def name(self) -> str:
        """Return the name of this custom emoji."""
        return self.obj_ref.custom_emoji.name

    @property
    def id(self) -> UUID:
        """Return the ID of this custom emoji."""
        return self.obj_ref.custom_emoji.id

    @property
    def url(self) -> str:
        """Return the URL of this custom emoji."""
        return self.obj_ref.custom_emoji.url

    def __repr__(self) -> str:
        return get_repr(self, desc=self.name)

    def __str__(self) -> str:
        return f':{self.name}:'



"""Handling the configuration for all adapters"""

import logging
import os
import re
import sys
from pathlib import Path
from typing import Any

import tomli
from pydantic import BaseModel, FilePath, field_validator
from pydantic_core.core_schema import ValidationInfo

_logger = logging.getLogger(__name__)

ENV_ULTIMATE_NOTION_CFG: str = 'ULTIMATE_NOTION_CONFIG'
"""Name of the environment variable to look up the path for the config"""
DEFAULT_ULTIMATE_NOTION_CFG_PATH: str = '.ultimate-notion/config.toml'
"""Default path within $HOME to the configuration file of Ultimate Notion"""
ENV_NOTION_TOKEN = 'NOTION_TOKEN'  # same as in `notion-sdk-py` package
"""Name of the environment variable to look up the Notion token"""
ENV_ULTIMATE_NOTION_DEBUG = 'ULTIMATE_NOTION_DEBUG'

DEFAULT_CFG = f"""\
# Configuration for Ultimate Notion
#
# * Non-absolute paths are always relative to the directory of this file.
# * You can use environment variables in the format ${{env:VAR_NAME}} or ${{env:VAR_NAME|DEFAULT_VALUE}}.

[ultimate_notion]
sync_state_dir = "sync_states"
debug = "${{env:{ENV_ULTIMATE_NOTION_DEBUG}|false}}"
token = "${{env:{ENV_NOTION_TOKEN}}}"

[google]
client_secret_json = "client_secret.json"
token_json = "token.json"
"""


class UNOCfg(BaseModel):
    """Configuration related to Ultimate Notion itself."""

    token: str | None = None
    debug: bool = False
    sync_state_dir: Path = Path('sync_states')
    cfg_path: FilePath  # will be set automatically


class GoogleCfg(BaseModel):
    """Configuration related to the Google API."""

    client_secret_json: Path | None = None
    token_json: Path | None = None


class Config(BaseModel):
    """Main configuration object."""

    ultimate_notion: UNOCfg
    google: GoogleCfg | None = None

    @field_validator('ultimate_notion')
    @classmethod
    def uno_convert_path(cls, value: UNOCfg, info: ValidationInfo) -> UNOCfg:
        if not value.sync_state_dir.is_absolute():
            value.sync_state_dir = value.cfg_path.parent / value.sync_state_dir
            if not value.sync_state_dir.exists():
                value.sync_state_dir.mkdir(parents=True)
        return value

    @field_validator('google')
    @classmethod
    def google_convert_path(cls, value: GoogleCfg | None, info: ValidationInfo) -> GoogleCfg | None:
        def make_rel_path_abs(entry: FilePath | None) -> FilePath | None:
            if entry is not None and not entry.is_absolute():
                cfg_path: Path = info.data['ultimate_notion'].cfg_path
                entry = cfg_path.parent / entry
            return entry

        if value is not None:
            value.client_secret_json = make_rel_path_abs(value.client_secret_json)
            value.token_json = make_rel_path_abs(value.token_json)

        return value


def get_cfg_file() -> Path:
    """Determines the path of the config file."""
    path_str = os.environ.get(ENV_ULTIMATE_NOTION_CFG, None)
    path = Path.home() / Path(DEFAULT_ULTIMATE_NOTION_CFG_PATH) if path_str is None else Path(path_str)
    return path


def resolve_env_value(value: str) -> str:
    """Resolves environment variable values in the format ${env:VAR_NAME|DEFAULT_VALUE}."""
    match = re.match(r'\${env:(\w+)(?:\|(.*))?}$', value)
    if match:
        var_name = match.group(1)
        default_value = match.group(2)
        return os.environ.get(var_name, default_value)
    return value


def get_cfg() -> Config:
    """Returns the configuration as an object."""
    cfg_path = get_cfg_file()
    _logger.info(f'Loading configuration from path `{cfg_path}`.')
    with open(cfg_path, 'rb') as fh:
        # ToDo: Use `tomllib` when Python 3.11 is the minimum version
        cfg_dict = tomli.load(fh)

    # Resolve environment variable values in the configuration
    def resolve_values(data: Any) -> None:
        if isinstance(data, dict):
            for key, value in data.items():
                if isinstance(value, dict):
                    resolve_values(value)
                elif isinstance(value, str):
                    data[key] = resolve_env_value(value)

    resolve_values(cfg_dict)

    # add config path to later resolve relative paths of config values
    if 'ultimate_notion' not in cfg_dict:
        msg = f'The configuration file {cfg_path} does not contain a section [ultimate-notion].'
        raise RuntimeError(msg)

    cfg_dict['ultimate_notion']['cfg_path'] = cfg_path
    return Config.model_validate(cfg_dict)


def get_or_create_cfg() -> Config:
    """Returns the configuration as an object or creates it if it doesn't exist yet."""
    cfg_path = get_cfg_file()
    if not cfg_path.exists():
        cfg_path.parent.mkdir(parents=True, exist_ok=True)
        cfg_path.write_text(DEFAULT_CFG)
    return get_cfg()


def activate_debug_mode() -> None:
    """Activates debug mode by setting up logging and notifying the user."""

    from ultimate_notion import __version__  # noqa: PLC0415

    if not logging.getLogger().hasHandlers():
        logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)

    logger = logging.getLogger(__package__)
    logger.setLevel(logging.DEBUG)

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug(f'Ultimate Notion {__version__} is running in debug mode.')



"""Base classes for working with the Notion API.


!!! warning

   To have proper type checking and inference, covariant type parameters are sometimes used even though
   they may violate the LSP (Liskov Substitution Principle) in some cases. This was done to improve the usability
   and flexibility of the API, allowing for more intuitive code while still maintaining type safety.
   Especially in this complex class hierarchy, Pydantic, which is heavily used in the API, relies on these
   covariant type parameters to function correctly. As an alternative approach, Protocol classes were tried,
   but they introduced their own complexities and limitations.
"""

from __future__ import annotations

import builtins
import logging
import re
from datetime import datetime
from typing import TYPE_CHECKING, Any, ClassVar, Generic, NoReturn, cast, overload
from uuid import UUID

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    SerializeAsAny,
    ValidatorFunctionWrapHandler,
    field_validator,
    model_validator,
)
from typing_extensions import Self, TypeIs, TypeVar

from ultimate_notion.errors import UnsetError
from ultimate_notion.obj_api.enums import Color
from ultimate_notion.utils import is_stable_release, pydantic_apply

if TYPE_CHECKING:
    from ultimate_notion.obj_api.objects import ParentRef, UserRef


_logger = logging.getLogger(__name__)


BASE_URL_PATTERN = r'https://(www.)?notion.so/'
UUID_PATTERN = r'[0-9a-f]{8}-?[0-9a-f]{4}-?[0-9a-f]{4}-?[0-9a-f]{4}-?[0-9a-f]{12}'

UUID_RE = re.compile(rf'^(?P<id>{UUID_PATTERN})$')

PAGE_URL_SHORT_RE = re.compile(
    rf"""^
      {BASE_URL_PATTERN}
      (?P<page_id>{UUID_PATTERN})
    $""",
    flags=re.IGNORECASE | re.VERBOSE,
)

PAGE_URL_LONG_RE = re.compile(
    rf"""^
      {BASE_URL_PATTERN}
      (?P<title>.*)-
      (?P<page_id>{UUID_PATTERN})
    $""",
    flags=re.IGNORECASE | re.VERBOSE,
)

BLOCK_URL_LONG_RE = re.compile(
    rf"""^
      {BASE_URL_PATTERN}
      (?P<username>.*)/
      (?P<title>.*)-
      (?P<page_id>{UUID_PATTERN})
      \#(?P<block_id>{UUID_PATTERN})
    $""",
    flags=re.IGNORECASE | re.VERBOSE,
)


T = TypeVar('T')  # ToDo: use new syntax in Python 3.12 and consider using default = in Python 3.13+


class UnsetType(BaseModel):
    """Sentinel type for missing default values.

    A sentinel type for indicating that a value is unset or missing for cases when `None` has another meaning.
    In the Notion API, `None` is also used to delete properties, so a way to represent "no value" explicitly is needed.
    """

    __slots__ = ()

    def __new__(cls, *args: Any, **kwargs: Any) -> UnsetType:
        # Ensure only one instance is created, allowing `is` to work correctly
        if not hasattr(cls, '_instance'):
            cls._instance = super().__new__(cls)
        return cls._instance

    # Marker field that survives serialization to identify unset values
    unset_marker: bool = Field(default=True, exclude=False)

    def __repr__(self) -> str:
        return 'Unset'

    def __hash__(self) -> int:
        return hash(UnsetType)


Unset: UnsetType = UnsetType()


def is_unset(v: Any) -> TypeIs[UnsetType]:
    """Check if the given value is unset."""
    if isinstance(v, UnsetType):
        return True
    elif v == Unset.model_dump(mode='python'):  # v was serialized
        return True
    return False


@overload
def raise_unset(obj: UnsetType) -> NoReturn: ...
@overload
def raise_unset(obj: T) -> T: ...
def raise_unset(obj: T | UnsetType) -> T:
    """Raise an error if the object is unset."""
    if is_unset(obj):
        msg = 'Parameter is unset and was not yet initialized by the Notion API.'
        raise UnsetError(msg)
    return obj


def extract_id(text: str) -> str | None:
    """Examine the given text to find a valid Notion object ID."""

    m = UUID_RE.match(text)
    if m is not None:
        return m.group('id')

    m = PAGE_URL_LONG_RE.match(text)
    if m is not None:
        return m.group('page_id')

    m = PAGE_URL_SHORT_RE.match(text)
    if m is not None:
        return m.group('page_id')

    m = BLOCK_URL_LONG_RE.match(text)
    if m is not None:
        return m.group('block_id')

    return None


def _freeze(obj: Any) -> Any:
    """Make nested structures hashable & deterministic.

    !!! note

        We normalize the `color` field in some objects, since if it is `Unset`
        and has a default value of `Color.DEFAULT` in equality comparisons.
        We also ignore `id` fields here, e.g. in `Option` objects, since they
        are not relevant for equality and hashing.
    """
    if isinstance(obj, BaseModel):
        return _freeze(obj.model_dump(mode='python'))
    if isinstance(obj, dict):
        elems = []
        for k, v in sorted(obj.items()):
            if k == 'color' and is_unset(v):
                v = _normalize_color(v)
            elif k in {'id', 'option_ids'}:
                continue
            elems.append((k, _freeze(v)))
        return tuple(elems)
    if isinstance(obj, (list, tuple)):
        return tuple(_freeze(x) for x in obj)
    if isinstance(obj, set):
        return frozenset(_freeze(x) for x in obj)
    return obj


def _normalize_color(val: Color | UnsetType) -> Color:
    """Normalize unset colors to Color.DEFAULT.

    This is useful for ensuring consistent behavior when comparing objects. Notion API sometimes
    demands a color field to be `Unset` but then returns the default color, which can lead to
    unexpected behavior in equality checks and hashing.
    """
    if is_unset(val):
        return Color.DEFAULT
    return val


class GenericObject(BaseModel):
    """The base for all API objects."""

    model_config = ConfigDict(extra='ignore' if is_stable_release() else 'forbid')

    def __eq__(self, other: Any) -> bool:
        """Compare two objects for equality by comparing all their fields."""
        if not isinstance(other, GenericObject):
            return NotImplemented

        # _freeze is used to guarantee the consistency with __hash__.
        return bool(_freeze(self) == _freeze(other))

    def __hash__(self) -> int:
        """Compute a hash value for the object by hashing all its fields."""
        frozen = _freeze(self)
        return hash(frozen)

    @classmethod
    def _set_field_default(cls, name: str, default: str) -> None:
        """Modify the `BaseModel` field information for a specific class instance.

        This is necessary in particular for subclasses that change the default values
        of a model when defined. Notable examples are `TypedObject` and `NotionObject`.

        Args:
            name: the named attribute in the class
            default: the new default value for the named field
        """
        # Rebuild model to avoid UserWarning about shadowing an attribute in parent.
        # More details here: https://github.com/pydantic/pydantic/issues/6966
        field = cls.model_fields.get(name)
        if field is None:
            msg = f'No field of name {name} in {cls.__name__} found!'
            raise ValueError(msg)
        field.default = default
        field.validate_default = False
        cls.model_rebuild(force=True)

    # https://github.com/pydantic/pydantic/discussions/3139
    def update(self, **data: Any) -> None:
        """Update the internal attributes with new data in place."""

        new_obj_dct = self.model_dump()
        new_obj_dct.update(data)
        new_obj = self.model_validate(new_obj_dct)

        for k, v in new_obj.model_dump(exclude_unset=True).items():
            # exclude_unset avoids overwriting for instance known children that need to be retrieved separately
            _logger.debug('updating object data: %s => %s', k, v)
            setattr(self, k, getattr(new_obj, k))

    def serialize_for_api(self) -> dict[str, Any]:
        """Serialize the object for sending it to the Notion API."""

        def remove_unset(key: str, value: Any) -> Any:
            if is_unset(value):
                return None
            return value

        filtered_obj = pydantic_apply(self, remove_unset)
        # Notion API doesn't like "null" values
        return filtered_obj.model_dump(mode='json', exclude_none=True, by_alias=True)

    @classmethod
    def build(cls, *args: Any, **kwargs: Any) -> Self:
        """Use the standard constructur to build the instance. Will be overwritten for more complex types."""
        return cls(*args, **kwargs)


class UniqueObject(GenericObject):
    """A Notion object that has a unique ID.

    This is the base class for all Notion objects that have a unique identifier, i.e. `id`.

    !!! warning

        The `id` field is only set when the object is sent or retrieved from the API, not when created locally.
    """

    id: UUID | str | UnsetType = Field(union_mode='left_to_right', default=Unset)
    """`id` is an `UUID` if possible or a string (possibly not unique) depending on the object"""


class NotionObject(UniqueObject):
    """A top-level Notion API resource.

    Many objects in the Notion API follow a standard pattern with a `object` property, which
    defines the general object type, e.g. `page`, `database`, `user`, `block`, ...
    """

    object: str = Field(default=None)  # type: ignore # avoids mypy plugin errors as this is set in __init_subclass__
    """`object` is a string that identifies the general object type, e.g. `page`, `database`, `user`, `block`, ..."""

    request_id: UUID | None = None
    """`request_id` is a UUID that is used to track requests in the Notion API"""

    def __init_subclass__(cls, *, object: str | None = None, **kwargs: Any) -> None:  # noqa: A002
        super().__init_subclass__(**kwargs)

    @classmethod
    def __pydantic_init_subclass__(cls, *, object: str | None = None, **kwargs: Any) -> None:  # noqa: A002, PLW3201
        """Update `GenericObject` defaults for the named object.

        Needed since `model_fields` are not available during __init_subclass__
        See: https://github.com/pydantic/pydantic/issues/5369
        """
        super().__pydantic_init_subclass__(object=object, **kwargs)

        if object is not None:  # if None we inherit 'object' from the base class
            cls._set_field_default('object', default=object)

    @field_validator('object', mode='after')
    @classmethod
    def _verify_object_matches_expected(cls, val: str) -> str:
        """Make sure that the deserialized object matches the name in this class."""
        if (obj_field := cls.model_fields.get('object')) is not None:
            obj_attr = obj_field.default
        else:
            msg = 'Field `object` of Notion object is missing'
            raise ValueError(msg)

        if val != obj_attr:
            msg = f'Invalid object for {obj_attr} - {val}'
            raise ValueError(msg)

        return val


class NotionEntity(NotionObject):
    """A materialized entity, which was created by a user."""

    id: UUID | UnsetType = Unset
    parent: SerializeAsAny[ParentRef] | UnsetType = Unset

    created_time: datetime | UnsetType = Unset
    created_by: UserRef | UnsetType = Unset

    last_edited_time: datetime | UnsetType = Unset


TO_co = TypeVar('TO_co', covariant=True, default=Any)  # ToDo: use new syntax in Python 3.12


class TypedObject(GenericObject, Generic[TO_co]):
    """A type-referenced object.

    Many objects in the Notion API follow a standard pattern with a `type` property
    followed by additional data. These objects must specify a `type` attribute to
    ensure that the correct object is created.

    For example, this contains a nested 'detail' object:

        data = {
            type: "detail",
            ...
            detail: {
                ...
            }
        }
    """

    type: str = Field(default=None)  # type: ignore  # avoids mypy errors as this is set in __pydantic_init_subclass__
    """`type` is a string that identifies the specific object type, e.g. `heading_1`, `paragraph`, `equation`, ..."""
    _polymorphic_base: ClassVar[bool] = False
    _typemap: ClassVar[dict[str, builtins.type[TypedObject]]]

    def __init_subclass__(cls, *, type: str | None = None, polymorphic_base: bool = False, **kwargs: Any) -> None:  # noqa: A002
        super().__init_subclass__(**kwargs)
        cls._polymorphic_base = polymorphic_base

    @classmethod
    def __pydantic_init_subclass__(cls, *, type: str | None = None, **kwargs: Any) -> None:  # noqa: A002, PLW3201
        """Register the subtypes of the TypedObject subclass.

        This is needed since `model_fields` is not available during __init_subclass__.
        See: https://github.com/pydantic/pydantic/issues/5369
        """
        super().__pydantic_init_subclass__(type=type, **kwargs)
        type_name = cls.__name__ if type is None else type
        cls._register_type(type_name)

    @classmethod
    def _register_type(cls, name: str) -> None:
        """Register a specific class for the given 'type' name."""

        cls._set_field_default('type', default=name)

        # initialize a _typemap map for each direct child of TypedObject

        # this allows different class trees to have the same 'type' name
        # but point to a different object (e.g. the 'date' type may have
        # different implementations depending where it is used in the API)

        if not hasattr(cls, '_typemap'):
            cls._typemap = {}

        if name in cls._typemap:
            msg = f'Duplicate subtype for class - {name} :: {cls}'
            raise ValueError(msg)

        _logger.debug('registered new subtype: %s => %s', name, cls)
        cls._typemap[name] = cls

    @model_validator(mode='wrap')
    @classmethod
    def _resolve_type(cls, value: Any, handler: ValidatorFunctionWrapHandler) -> Any:
        """Instantiate the correct object based on the `type` field.

        Following this approach: https://github.com/pydantic/pydantic/discussions/7008
        Also the reason for `polymorphic_base` is explained there.
        """

        if isinstance(value, cls):
            return handler(value)

        if not cls._polymorphic_base:  # breaks the recursion
            return handler(value)

        if not isinstance(value, dict):
            msg = "Invalid 'data' object"
            raise ValueError(msg)

        if not hasattr(cls, '_typemap'):
            msg = f"Missing '_typemap' in {cls}"
            raise ValueError(msg)

        type_name = value.get('type')

        if type_name is None:
            _logger.warning(f'Missing type in data {value}. Most likely a User object without type')
            msg = f"Missing 'type' in data {value}"
            if value['object'] == 'user':
                type_name = 'unknown'  # for the unofficial type objects.UnknownUser
            else:
                raise ValueError(msg)

        sub_cls = cls._typemap.get(type_name)

        if sub_cls is None:
            msg = f'Unsupported sub-type: {type_name}'
            raise ValueError(msg)

        return sub_cls(**value)

    @property
    def value(self) -> TO_co:
        """Return the nested object."""
        return cast(TO_co, getattr(self, self.type))

    @value.setter
    def value(self, val: TO_co) -> None:  # type: ignore[misc]  # breaking covariance
        """Set the nested object."""
        # we are breaking covariance here but going down the Protocol way didn't work out
        # either due to many limititations, e.g. @runtime_checkable not working with inheritance, etc.
        # This is still the most programmatic way it seems without adding too much complexity.
        setattr(self, self.type, val)



"""Utilities for working with plain & rich texts in Notion."""

from __future__ import annotations

import datetime as dt
import re
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, Any, TypeGuard
from urllib.parse import urlparse

import pendulum as pnd
from typing_extensions import Self, TypeVar
from url_normalize import url_normalize

from ultimate_notion.core import Wrapper
from ultimate_notion.emoji import CustomEmoji
from ultimate_notion.markdown import render_md, rich_texts_to_markdown
from ultimate_notion.obj_api import objects as objs
from ultimate_notion.obj_api.core import Unset
from ultimate_notion.obj_api.enums import Color
from ultimate_notion.obj_api.objects import MAX_TEXT_OBJECT_SIZE
from ultimate_notion.user import User

if TYPE_CHECKING:
    from ultimate_notion.database import Database
    from ultimate_notion.page import Page


# ToDo: Use new syntax when requires-python >= 3.12
RTBO_co = TypeVar('RTBO_co', bound=objs.RichTextBaseObject, default=objs.RichTextBaseObject, covariant=True)


class RichTextBase(Wrapper[RTBO_co], wraps=objs.RichTextBaseObject):
    """Super class for text, equation and mentions of various kinds."""

    def __init__(self, *args: Any, href: str | None, **kwargs: Any) -> None:
        if href is not None:
            href = url_normalize(href)

        super().__init__(*args, href=href, **kwargs)

    @property
    def is_text(self) -> bool:
        return isinstance(self, RichText)

    @property
    def is_equation(self) -> bool:
        return isinstance(self, Math)

    @property
    def is_mention(self) -> bool:
        return isinstance(self, Mention)


class Math(RichTextBase[objs.EquationObject], wraps=objs.EquationObject):
    """An inline equation object.

    A LaTeX equation in inline mode, e.g. `$ \\mathrm{E=mc^2} $`, but without the `$` signs.

    !!! note

        Only used internally. Use the `math` function instead, to create proper `Text` object.
    """

    def __init__(
        self,
        expression: str,
        *,
        bold: bool = False,
        italic: bool = False,
        strikethrough: bool = False,
        code: bool = False,
        underline: bool = False,
        color: Color = Color.DEFAULT,
        href: str | None = None,
    ) -> None:
        annotations = objs.Annotations(
            bold=bold, italic=italic, strikethrough=strikethrough, code=code, underline=underline, color=color
        )
        super().__init__(expression, href=href, style=annotations)


def math(
    expression: str,
    *,
    bold: bool = False,
    italic: bool = False,
    strikethrough: bool = False,
    code: bool = False,
    underline: bool = False,
    color: Color = Color.DEFAULT,
) -> Text:
    """Create a Text that holds a formula."""
    math = Math(
        expression, bold=bold, italic=italic, strikethrough=strikethrough, code=code, underline=underline, color=color
    )
    return Text.wrap_obj_ref([math.obj_ref])


class Mention(RichTextBase[objs.MentionObject], wraps=objs.MentionObject):
    """A Mention object.

    !!! note

        Only used internally. Use the `mention` function instead, to create proper `RichText` object.
    """

    def __init__(
        self,
        target: User | Page | Database | CustomEmoji | objs.DateTimeOrRange,
        *,
        bold: bool = False,
        italic: bool = False,
        strikethrough: bool = False,
        code: bool = False,
        underline: bool = False,
        color: Color = Color.DEFAULT,
    ) -> None:
        annotations = objs.Annotations(
            bold=bold, italic=italic, strikethrough=strikethrough, code=code, underline=underline, color=color
        )
        if isinstance(target, dt.datetime | dt.date | pnd.Interval):
            self.obj_ref = objs.DateRange.build(target).build_mention(style=annotations)
        else:
            self.obj_ref = target.obj_ref.build_mention(style=annotations)

    @property
    def type(self) -> str:
        """Type of the mention, e.g. user, page, etc."""
        return self.obj_ref.mention.type


def mention(
    target: User | Page | Database | CustomEmoji | objs.DateTimeOrRange,
    *,
    bold: bool = False,
    italic: bool = False,
    strikethrough: bool = False,
    code: bool = False,
    underline: bool = False,
    color: Color = Color.DEFAULT,
) -> Text:
    """Create a Text that mentions another object."""
    mention = Mention(
        target, bold=bold, italic=italic, strikethrough=strikethrough, code=code, underline=underline, color=color
    )
    return Text.wrap_obj_ref([mention.obj_ref])


class RichText(RichTextBase[objs.TextObject], wraps=objs.TextObject):
    """A RichText object defining a formatted text fragment.

    !!! note

        Only used internally. Use `Text` instead.
    """

    def __init__(
        self,
        text: str,
        *,
        bold: bool = False,
        italic: bool = False,
        strikethrough: bool = False,
        code: bool = False,
        underline: bool = False,
        color: Color | None = None,
        href: str | None = None,
    ) -> None:
        if len(text) > MAX_TEXT_OBJECT_SIZE:
            msg = f'Text exceeds the maximum size of {MAX_TEXT_OBJECT_SIZE} characters. Use `Text` object instead!'
            raise ValueError(msg)

        color_or_unset = Unset if color is None else color
        annotations = objs.Annotations(
            bold=bold, italic=italic, strikethrough=strikethrough, code=code, underline=underline, color=color_or_unset
        )
        super().__init__(text, href=href, style=annotations)


class Text(str):
    """User-facing class holding several RichTextsBase objects.

    Rather use the constructor function `text` to create a `Text` object from a normal string with formatting.
    """

    _rich_texts: list[RichTextBase]

    def __init__(self, text: str) -> None:
        # note that super().__new__ (which is inherited) stores the plain text in the object for `str(self)`
        super().__init__()

        self._rich_texts: list[RichTextBase] = []
        if isinstance(text, Text):
            self._rich_texts = text._rich_texts
        else:
            for part in chunky(text):
                self._rich_texts.append(RichText(part))

    @property
    def rich_texts(self) -> tuple[RichTextBase, ...]:
        """Return the rich texts as immutable tuple."""
        return tuple(self._rich_texts)

    @classmethod
    def wrap_obj_ref(cls, obj_refs: list[objs.RichTextBaseObject] | None) -> Self:
        obj_refs = [] if obj_refs is None else obj_refs
        rich_texts = [RichTextBase.wrap_obj_ref(obj_ref) for obj_ref in obj_refs]
        plain_text = ''.join(text.obj_ref.plain_text for text in rich_texts if text)
        obj = cls(plain_text)
        obj._rich_texts = rich_texts
        return obj

    @property
    def obj_ref(self) -> list[objs.RichTextBaseObject]:
        return [elem.obj_ref for elem in self._rich_texts]

    @classmethod
    def from_markdown(cls, text: str) -> Text:
        """Create RichTextList by parsing the markdown."""
        # ToDo: Implement me!
        # ToDo: Handle Equations and Mentions here accordingly
        raise NotImplementedError

    def to_markdown(self) -> str:
        """Convert the list of RichText objects to markdown."""
        return rich_texts_to_markdown(self._rich_texts)

    @classmethod
    def from_plain_text(cls, text: str) -> Text:
        """Create RichTextList from plain text.

        This method is a more explicit alias for the default constructor.
        """
        return cls(text)

    def to_plain_text(self) -> str:
        """Return rich text as plain text

        This method is a more explicit variant then just using the object.
        """
        return str(self)

    @property
    def mentions(self) -> tuple[Mention, ...]:
        """Return all mentions in the text."""

        def is_mention(rt: RichTextBase) -> TypeGuard[Mention]:
            return rt.is_mention

        return tuple(rt for rt in self._rich_texts if is_mention(rt))

    def to_html(self) -> str:
        """Return rich text as HTML."""
        return render_md(self.to_markdown())

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by JupyterLab automatically to display this text."""
        return self.to_html()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str | Text):
            return str(self) == str(other)
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash(str(self))

    def __add__(self, other: str) -> Text:
        match other:
            case CustomEmoji():  # got to wrap CusomEmoji in a Mention
                return Text.wrap_obj_ref([*self.obj_ref, *mention(other).obj_ref])
            case str():
                return Text.wrap_obj_ref([*self.obj_ref, *Text(other).obj_ref])
            case _:
                msg = f'Cannot concatenate {type(other)} to construct a RichText object.'
                raise RuntimeError(msg)


def text(
    text: str,
    *,
    bold: bool = False,
    italic: bool = False,
    strikethrough: bool = False,
    code: bool = False,
    underline: bool = False,
    color: Color | None = None,
    href: str | None = None,
) -> Text:
    """Create a rich text Text object from a normal string with formatting.

    !!! warning

        If a `Text` object is passed, the original formatting will be lost!
    """
    if isinstance(text, Text):
        text = text.to_plain_text()

    return Text.wrap_obj_ref(
        [
            RichText(
                part,
                bold=bold,
                italic=italic,
                strikethrough=strikethrough,
                code=code,
                underline=underline,
                color=color,
                href=href,
            ).obj_ref
            for part in chunky(text)
        ]
    )


def chunky(text: str, length: int = MAX_TEXT_OBJECT_SIZE) -> Iterator[str]:
    """Break the given `text` into chunks of at most `length` size."""
    return (text[idx : idx + length] for idx in range(0, len(text), length))


def python_identifier(string: str) -> str:
    """Make a valid Python identifier.

    This will remove any leading characters that are not valid and change all
    invalid interior sequences to underscore.

    Attention: This may result in an empty string!
    """
    s = re.sub(r'[^0-9a-zA-Z_]+', '_', string)
    s = re.sub(r'^[^a-zA-Z]+', '', s)
    return s.rstrip('_')


def snake_case(string: str) -> str:
    """Make a Python identifier in snake_case.

    Attention: This may result in an empty string!
    """
    return python_identifier(string).lower()


def camel_case(string: str) -> str:
    """Make a Python identifier in CamelCase.

    Attention: This may result in an empty string and a CamelCase sting will be capitalized!
    """
    return ''.join([elem.capitalize() for elem in snake_case(string).split('_')])


def decapitalize(string: str) -> str:
    """Inverse of `capitalize`."""
    if not string:
        return string
    return string[0].lower() + string[1:]


def is_url(string: str) -> bool:
    """Check if a string is a valid URL."""
    try:
        result = urlparse(string)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def html_img(url: str, size: float) -> str:
    """Create a img tag in HTML."""
    return f'<img src="{url}" style="height:{size:.2f}em">'


def join(texts: Sequence[str], *, delim: str = ' ') -> Text:
    """Join multiple str objects, including Text, into a single Text object with a given delimeter."""
    if len(texts) == 0:
        return Text.wrap_obj_ref([])

    if isinstance(delim, str):
        delim_objs = Text(delim).obj_ref

    all_objs = Text(texts[0]).obj_ref
    for text in texts[1:]:
        all_objs.extend([*delim_objs, *Text(text).obj_ref])

    return Text.wrap_obj_ref(all_objs)



"""Objects representing a database schema.

Properties are used when

1. a database with a specific schema is [created],
2. a database with a schema is [retrieved].

Unfortunately, the way a schema is defined in case of 1. and 2. is not consistent.
In case 1., the property name is only defined as a key while in case 2. it is additionally
defined as `name` attribute of the property object. We treat these two cases the same way
when constructing the property objects. For this reason `name` is `Unset` by default.

[created]: https://developers.notion.com/reference/property-schema-object
[retrieved]: https://developers.notion.com/reference/property-object
"""

from __future__ import annotations

from uuid import UUID

from pydantic import Field, field_validator
from typing_extensions import TypeVar

from ultimate_notion.obj_api.core import GenericObject, TypedObject, Unset, UnsetType
from ultimate_notion.obj_api.enums import AggFunc, NumberFormat
from ultimate_notion.obj_api.objects import SelectGroup, SelectOption

# ToDo: Use new syntax when requires-python >= 3.12
GO_co = TypeVar('GO_co', bound=GenericObject, covariant=True, default=GenericObject)


class Property(TypedObject[GO_co], polymorphic_base=True):
    """Base class for Notion property objects."""

    id: str | UnsetType = Unset
    name: str | UnsetType = Unset  # Unset when creating a database property schema
    description: str | None = None


class TitleTypeData(GenericObject):
    """Type data for `Title`."""


class Title(Property[TitleTypeData], type='title'):
    """Defines the title configuration for a database property."""

    title: TitleTypeData = Field(default_factory=TitleTypeData)


class RichTextTypeData(GenericObject):
    """Type data for `RichText`."""


class RichText(Property[RichTextTypeData], type='rich_text'):
    """Defines the rich text configuration for a database property."""

    rich_text: RichTextTypeData = Field(default_factory=RichTextTypeData)


class NumberTypeData(GenericObject):
    """Type data for `Number`."""

    format: NumberFormat = NumberFormat.NUMBER

    # leads to better error messages, see
    # https://github.com/pydantic/pydantic/issues/355
    @field_validator('format')
    @classmethod
    def validate_enum_field(cls, field: str) -> NumberFormat:
        return NumberFormat(field)


class Number(Property[NumberTypeData], type='number'):
    """Defines the number configuration for a database property."""

    number: NumberTypeData = Field(default_factory=NumberTypeData)

    @classmethod
    def build(cls, format: NumberFormat) -> Number:  # noqa: A002
        """Create a `Number` object with the expected format."""
        return cls.model_construct(number=NumberTypeData(format=format))


class SelectTypeData(GenericObject):
    """Type data for `Select`."""

    options: list[SelectOption] = Field(default_factory=list)


class Select(Property[SelectTypeData], type='select'):
    """Defines the select configuration for a database property."""

    select: SelectTypeData = Field(default_factory=SelectTypeData)

    @classmethod
    def build(cls, options: list[SelectOption]) -> Select:
        """Create a `Select` object from the list of `SelectOption`'s."""
        return cls.model_construct(select=SelectTypeData(options=options))


class MultiSelectTypeData(GenericObject):
    """Type data for `MultiSelect`."""

    options: list[SelectOption] = Field(default_factory=list)


class MultiSelect(Property[MultiSelectTypeData], type='multi_select'):
    """Defines the multi-select configuration for a database property."""

    multi_select: MultiSelectTypeData = Field(default_factory=MultiSelectTypeData)

    @classmethod
    def build(cls, options: list[SelectOption]) -> MultiSelect:
        """Create a `Select` object from the list of `SelectOption`'s."""
        return cls.model_construct(multi_select=MultiSelectTypeData(options=options))


class StatusTypeData(GenericObject):
    """Type data for `Status`."""

    options: list[SelectOption] = Field(default_factory=list)
    groups: list[SelectGroup] = Field(default_factory=list)


class Status(Property[StatusTypeData], type='status'):
    """Defines the status configuration for a database property."""

    status: StatusTypeData = Field(default_factory=StatusTypeData)

    @classmethod
    def build(cls, options: list[SelectOption], groups: list[SelectGroup]) -> Status:
        """Create a `Status` object from the list of `SelectOption`'s.

        !!! warning

            While a Status property can be built, it can only be used to
            check a schema, not to create a database having such a property.
        """
        return cls.model_construct(status=StatusTypeData(options=options, groups=groups))


class DateTypeData(GenericObject):
    """Type data for `Date`."""


class Date(Property[DateTypeData], type='date'):
    """Defines the date configuration for a database property."""

    date: DateTypeData = Field(default_factory=DateTypeData)


class PeopleTypeData(GenericObject):
    """Type data for `People`."""


class People(Property[PeopleTypeData], type='people'):
    """Defines the people configuration for a database property."""

    people: PeopleTypeData = Field(default_factory=PeopleTypeData)


class FilesTypeData(GenericObject):
    """Type data for `Files`."""


class Files(Property[FilesTypeData], type='files'):
    """Defines the files configuration for a database property."""

    files: FilesTypeData = Field(default_factory=FilesTypeData)


class CheckboxTypeData(GenericObject):
    """Type data for `Checkbox`."""


class Checkbox(Property[CheckboxTypeData], type='checkbox'):
    """Defines the checkbox configuration for a database property."""

    checkbox: CheckboxTypeData = Field(default_factory=CheckboxTypeData)


class EmailTypeData(GenericObject):
    """Type data for `Email`."""


class Email(Property[EmailTypeData], type='email'):
    """Defines the email configuration for a database property."""

    email: EmailTypeData = Field(default_factory=EmailTypeData)


class URLTypeData(GenericObject):
    """Type data for `URL`."""


class URL(Property[URLTypeData], type='url'):
    """Defines the URL configuration for a database property."""

    url: URLTypeData = Field(default_factory=URLTypeData)


class PhoneNumberTypeData(GenericObject):
    """Type data for `PhoneNumber`."""


class PhoneNumber(Property[PhoneNumberTypeData], type='phone_number'):
    """Defines the phone number configuration for a database property."""

    phone_number: PhoneNumberTypeData = Field(default_factory=PhoneNumberTypeData)


class FormulaTypeData(GenericObject):
    """Type data for `Formula`."""

    expression: str

    def __eq__(self, other: object) -> bool:
        """Compare Formula objects by all attributes except id."""
        # Sadly, expressions are changed by the Notion API, e.g. 'prop("Name")' in our request becomes
        # {{notion:block_property:title:00000000-0000-0000-0000-000000000000:5a65efbd-bfb2-4ebe-bb5d-ac95c98fb252}}
        # ToDo: Implement a way to compare these expressions, which is possible in principle.
        return isinstance(other, FormulaTypeData)

    def __hash__(self) -> int:
        """Return a hash of the Formula TypeData.

        Since __eq__ always returns True due to API expression transformation,
        we use a constant hash to maintain consistency with equality.
        """
        return hash(self.__class__)


class Formula(Property[FormulaTypeData], type='formula'):
    """Defines the formula configuration for a database property."""

    formula: FormulaTypeData

    @classmethod
    def build(cls, formula: str) -> Formula:
        return cls.model_construct(formula=FormulaTypeData(expression=formula))


class PropertyRelation(TypedObject[GO_co], polymorphic_base=True):
    """Defines common configuration for a property relation."""

    database_id: UUID
    data_source_id: str | None = None  # 2025-09-03 update: https://developers.notion.com/docs/upgrade-guide-2025-09-03

    def __eq__(self, other: object) -> bool:
        # ToDo: Consider data_source_id when Update 2025-09-03 is implemented.
        if not isinstance(other, PropertyRelation):
            return NotImplemented
        return self.database_id == other.database_id

    def __hash__(self) -> int:
        return hash(self.database_id)


class SinglePropertyRelationTypeData(GenericObject):
    """Type data for `SinglePropertyRelation`."""


class SinglePropertyRelation(PropertyRelation[SinglePropertyRelationTypeData], type='single_property'):
    """Defines a one-way relation configuration for a database property."""

    single_property: SinglePropertyRelationTypeData = Field(default_factory=SinglePropertyRelationTypeData)

    @classmethod
    def build_relation(cls, dbref: UUID) -> Relation:
        """Create a `single_property` relation using the target database reference.

        `dbref` must be either a string or UUID.
        """
        rel = SinglePropertyRelation.model_construct(database_id=dbref)
        return Relation.model_construct(relation=rel)


class DualPropertyRelationTypeData(GenericObject):
    """Type data for `DualPropertyRelation`."""

    synced_property_name: str | UnsetType = Unset
    synced_property_id: str | UnsetType = Unset

    def __eq__(self, other: object) -> bool:
        """Compare DualPropertyRelation objects by all attributes except id."""
        if not isinstance(other, DualPropertyRelationTypeData):
            return NotImplemented
        # we skip the id as this is set by the Notion API.
        return self.synced_property_name == other.synced_property_name

    def __hash__(self) -> int:
        """Return a hash of the DualPropertyRelation TypeData based on synced_property_name."""
        return hash(self.synced_property_name)


class DualPropertyRelation(PropertyRelation[DualPropertyRelationTypeData], type='dual_property'):
    """Defines a two-way relation configuration for a database property.

    If a two-way relation property X relates to Y then the two-way relation property Y relates to X.
    """

    dual_property: DualPropertyRelationTypeData = Field(default_factory=DualPropertyRelationTypeData)

    @classmethod
    def build_relation(cls, dbref: UUID) -> Relation:
        """Create a `dual_property` relation using the target database reference.

        `dbref` must be either a string or UUID.
        """
        # No `synced_property_name` is set since it will be ignored by the Notion API.
        # Thus we first get the default two-way relation name, which we gonna change later.
        # See: https://developers.notion.com/reference/property-schema-object#dual-property-relation-configuration
        rel = DualPropertyRelation.model_construct(database_id=dbref)
        return Relation.model_construct(relation=rel)


class Relation(Property[SinglePropertyRelation | DualPropertyRelation], type='relation'):
    """Defines the relation configuration for a database property."""

    relation: SinglePropertyRelation | DualPropertyRelation


class RollupTypeData(GenericObject):
    """Type data for `Rollup`."""

    function: AggFunc = AggFunc.COUNT_ALL

    relation_property_name: str
    relation_property_id: str | UnsetType = Unset

    rollup_property_name: str
    rollup_property_id: str | UnsetType = Unset

    # leads to better error messages, see
    # https://github.com/pydantic/pydantic/issues/355
    @field_validator('function')
    @classmethod
    def validate_enum_field(cls, field: str) -> AggFunc:
        return AggFunc(field)

    def __eq__(self, value: object) -> bool:
        """Compare Rollup objects by all attributes except id."""
        if not isinstance(value, RollupTypeData):
            return NotImplemented
        return (
            self.function == value.function
            and self.relation_property_name == value.relation_property_name
            and self.rollup_property_name == value.rollup_property_name
        )

    def __hash__(self) -> int:
        """Return a hash of the Rollup TypeData based on function and property names."""
        return hash((self.function, self.relation_property_name, self.rollup_property_name))


class Rollup(Property[RollupTypeData], type='rollup'):
    """Defines the rollup configuration for a database property."""

    rollup: RollupTypeData

    @classmethod
    def build(cls, relation: str, property: str, function: AggFunc) -> Rollup:  # noqa: A002
        return Rollup.model_construct(
            rollup=RollupTypeData(function=function, relation_property_name=relation, rollup_property_name=property)
        )


class CreatedTimeTypeData(GenericObject):
    """Type data for `CreatedTime`."""


class CreatedTime(Property[CreatedTimeTypeData], type='created_time'):
    """Defines the created-time configuration for a database property."""

    created_time: CreatedTimeTypeData = Field(default_factory=CreatedTimeTypeData)


class CreatedByTypeData(GenericObject):
    """Type data for `CreatedBy`."""


class CreatedBy(Property[CreatedByTypeData], type='created_by'):
    """Defines the created-by configuration for a database property."""

    created_by: CreatedByTypeData = Field(default_factory=CreatedByTypeData)


class LastEditedByTypeData(GenericObject):
    """Type data for `LastEditedBy`."""


class LastEditedBy(Property[LastEditedByTypeData], type='last_edited_by'):
    """Defines the last-edited-by configuration for a database property."""

    last_edited_by: LastEditedByTypeData = Field(default_factory=LastEditedByTypeData)


class LastEditedTimeTypeData(GenericObject):
    """Type data for `LastEditedTime`."""


class LastEditedTime(Property[LastEditedTimeTypeData], type='last_edited_time'):
    """Defines the last-edited-time configuration for a database property."""

    last_edited_time: LastEditedTimeTypeData = Field(default_factory=LastEditedTimeTypeData)


class UniqueIDTypeData(GenericObject):
    """Type data for `UniqueID`."""

    prefix: str | None = None


class UniqueID(Property[UniqueIDTypeData], type='unique_id'):
    """Unique ID database property."""

    unique_id: UniqueIDTypeData = Field(default_factory=UniqueIDTypeData)

    @classmethod
    def build(cls, prefix: str | None = None) -> UniqueID:
        return UniqueID.model_construct(unique_id=UniqueIDTypeData(prefix=prefix))


class VerificationTypeData(GenericObject):
    """Type data for `Verification`."""


class Verification(Property[VerificationTypeData], type='verification'):
    """Verfication database property of Wiki databases."""

    verification: VerificationTypeData = Field(default_factory=VerificationTypeData)


class ButtonTypeData(GenericObject):
    """Type data for `Button`."""


class Button(Property[ButtonTypeData], type='button'):
    """Button database property."""

    button: ButtonTypeData = Field(default_factory=ButtonTypeData)


class RenameProp(GenericObject):
    """Property to rename a property during a database update."""

    name: str



"""Property values of a page in Notion directly mapped to Python objects.

`PropertyValue` objects are directly returned from the Notion API using the [retrieve a page] endpoint.
They are used to represent the values of a page's properties. In contrast the `PropertyItem` objects
are returned from the [retrieve a page property item] endpoint and differ from the `PropertyValue` objects
by having a field `object = 'property_item'`. So they are considered *proper* objects in the Notion API
instead of just types like the `PropertyValue` objects.

[retrieve a page]: https://developers.notion.com/reference/retrieve-a-page
[retrieve a page property item]: https://developers.notion.com/reference/retrieve-a-page-property
"""

from __future__ import annotations

import datetime as dt
from abc import ABC
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, cast

from pydantic import Field, SerializeAsAny, field_validator, model_serializer
from typing_extensions import Self

from ultimate_notion.obj_api.core import GenericObject, NotionObject, Unset, UnsetType
from ultimate_notion.obj_api.enums import FormulaType, RollupType, VState
from ultimate_notion.obj_api.objects import (
    DateRange,
    FileObject,
    ObjectRef,
    RichTextBaseObject,
    TypedObject,
    User,
)
from ultimate_notion.obj_api.schema import AggFunc, SelectOption
from ultimate_notion.utils import DateTimeOrRange

if TYPE_CHECKING:
    from ultimate_notion.obj_api.blocks import Page

MAX_ITEMS_PER_PROPERTY = 25
"""Maximum number of items rertrieved per property.

Only a certain number of items for each property are retrieved by default.
The high-level API will retrieve the rest on demand automatically.

Source: https://developers.notion.com/reference/retrieve-a-page
"""

###################
# Property Values #
###################


class PropertyValue(TypedObject[Any], polymorphic_base=True):
    """Base class for Notion property values."""

    id: str | UnsetType = Unset
    _is_retrieved: bool = False  # fetched separately as property item from the server

    @classmethod
    def build(cls, value: Any) -> Self:
        """Build the property value from given value, e.g. native Python or nested type.

        In practice, this is like calling __init__ with the corresponding keyword.
        """
        return cast(Self, cls.model_construct(**{cls.model_fields['type'].get_default(): value}))

    def serialize_for_api(self) -> dict[str, Any]:
        """Serialize the object for sending it to the Notion API."""
        dump_dct = super().serialize_for_api()
        dump_dct.setdefault(dump_dct['type'], None)
        return dump_dct


class Title(PropertyValue, type='title'):
    """Notion title type."""

    title: list[SerializeAsAny[RichTextBaseObject]] = Field(default_factory=list)


class RichText(PropertyValue, type='rich_text'):
    """Notion rich text type."""

    rich_text: list[SerializeAsAny[RichTextBaseObject]] = Field(default_factory=list)


class Number(PropertyValue, type='number'):
    """Simple number type."""

    number: float | int | None = None


class Checkbox(PropertyValue, type='checkbox'):
    """Simple checkbox type; represented as a boolean."""

    checkbox: bool | None = None


class Date(PropertyValue, type='date'):
    """Notion complex date type - may include timestamp and/or be a date range."""

    date: DateRange | None = None

    @classmethod
    def build(cls, dt_spec: str | DateTimeOrRange) -> Self:
        """Create a new Date from the native values."""
        return cast(Self, cls.model_construct(date=DateRange.build(dt_spec)))


class Status(PropertyValue, type='status'):
    """Notion status property."""

    status: SelectOption | None = None


class Select(PropertyValue, type='select'):
    """Notion select type."""

    select: SelectOption | None = None


class MultiSelect(PropertyValue, type='multi_select'):
    """Notion multi-select type."""

    multi_select: list[SelectOption] = Field(default_factory=list)


class People(PropertyValue, type='people'):
    """Notion people type."""

    people: list[User] = Field(default_factory=list)

    # Custom serializer as we receive various UserRef subtypes but need to pass
    # a proper UserRef to the Notion API. Notion API is just so inconsistent!
    @model_serializer
    def serialize(self) -> dict[str, Any]:
        return {'people': [{'id': user.id, 'object': user.object} for user in self.people], 'type': 'people'}


class URL(PropertyValue, type='url'):
    """Notion URL type."""

    url: str | None = None


class Email(PropertyValue, type='email'):
    """Notion email type."""

    email: str | None = None


class PhoneNumber(PropertyValue, type='phone_number'):
    """Notion phone type."""

    phone_number: str | None = None


class Files(PropertyValue, type='files'):
    """Notion files type."""

    files: list[SerializeAsAny[FileObject]] = Field(default_factory=list)


class FormulaResult(TypedObject, ABC, polymorphic_base=True):
    """A Notion formula result.

    This object contains the result of the expression in the database properties.
    """


class StringFormula(FormulaResult, type=FormulaType.STRING.value):
    """A Notion string formula result."""

    string: str | None = None


class NumberFormula(FormulaResult, type=FormulaType.NUMBER.value):
    """A Notion number formula result."""

    number: float | int | None = None


class DateFormula(FormulaResult, type=FormulaType.DATE.value):
    """A Notion date formula result."""

    date: DateRange | None = None


class BooleanFormula(FormulaResult, type=FormulaType.BOOLEAN.value):
    """A Notion boolean formula result."""

    boolean: bool | None = None


class Formula(PropertyValue, type='formula'):
    """A Notion formula property value."""

    formula: SerializeAsAny[FormulaResult] | None = None


class Relation(PropertyValue, type='relation'):
    """A Notion relation property value."""

    relation: list[ObjectRef] = Field(default_factory=list)
    has_more: bool = False

    @classmethod
    def build(cls, pages: Sequence[Page]) -> Self:
        """Return a `Relation` property with the specified pages."""
        return cast(Self, cls.model_construct(relation=[ObjectRef.build(page) for page in pages]))


class RollupObject(TypedObject, ABC, polymorphic_base=True):
    """A Notion rollup property value."""

    function: AggFunc | None = None


class RollupNumber(RollupObject, type=RollupType.NUMBER.value):
    """A Notion rollup number property value."""

    number: float | int | None = None


class RollupDate(RollupObject, type=RollupType.DATE.value):
    """A Notion rollup date property value."""

    date: DateRange | None = None


class RollupArray(RollupObject, type=RollupType.ARRAY.value):
    """A Notion rollup array property value."""

    array: list[SerializeAsAny[PropertyValue]]


class RollupIncomplete(RollupObject, type=RollupType.INCOMPLETE.value):
    """A Notion incomplete rollup property value."""

    class TypeData(GenericObject): ...

    incomplete: TypeData


class RollupUnsupported(RollupObject, type=RollupType.UNSUPPORTED.value):
    """A Notion unsupported rollup property value."""

    class TypeData(GenericObject): ...

    unsupported: TypeData


class Rollup(PropertyValue, type='rollup'):
    """A Notion rollup property value."""

    rollup: SerializeAsAny[RollupObject] | None = None


class CreatedTime(PropertyValue, type='created_time'):
    """A Notion created-time property value."""

    created_time: dt.datetime


class CreatedBy(PropertyValue, type='created_by'):
    """A Notion created-by property value."""

    created_by: SerializeAsAny[User]


class LastEditedTime(PropertyValue, type='last_edited_time'):
    """A Notion last-edited-time property value."""

    last_edited_time: dt.datetime


class LastEditedBy(PropertyValue, type='last_edited_by'):
    """A Notion last-edited-by property value."""

    last_edited_by: SerializeAsAny[User]


class UniqueID(PropertyValue, type='unique_id'):
    """A Notion unique-id property value."""

    class TypeData(GenericObject):
        number: int = 0
        prefix: str | None = None

    unique_id: TypeData = Field(default_factory=TypeData)


class Verification(PropertyValue, type='verification'):
    """A Notion verification property value."""

    class TypeData(GenericObject):
        state: VState = VState.UNVERIFIED
        verified_by: SerializeAsAny[User] | None = None
        date: dt.datetime | None = None

        # leads to better error messages, see
        # https://github.com/pydantic/pydantic/issues/355
        @field_validator('state')
        @classmethod
        def validate_enum_field(cls, field: str) -> VState:
            return VState(field)

    verification: TypeData = Field(default_factory=TypeData)


class Button(PropertyValue, type='button'):
    """A Notion button property value."""

    class TypeData(GenericObject): ...

    button: TypeData = Field(default_factory=TypeData)


##################
# Property Items #
##################


class PropertyItem(NotionObject, TypedObject, polymorphic_base=True, object='property_item'):
    """A `PropertyItem` returned by the Notion API.

    Basic property items have a similar schema to corresponding property values.

    Notion-API: https://developers.notion.com/reference/property-item-object
    """

    id: str


class TitlePropertyItem(PropertyItem, type='title'):
    """A `PropertyItem` returned by the Notion API containing the `Title` property."""

    # According to the Notion API docs, this should be a list of rich text objects.
    # Nevertheless, the Notion API returns a single rich text object
    title: SerializeAsAny[RichTextBaseObject]


class RichTextPropertyItem(PropertyItem, type='rich_text'):
    """A `PropertyItem` returned by the Notion API containing the `RichText` property."""

    # According to the Notion API docs, this should be a list of rich text objects.
    # Nevertheless, the Notion API returns a single rich text object
    rich_text: SerializeAsAny[RichTextBaseObject]


class NumberPropertyItem(PropertyItem, Number, type='number'):
    """A `PropertyItem` returned by the Notion API containing the `Number` property."""


class CheckboxPropertyItem(PropertyItem, Checkbox, type='checkbox'):
    """A `PropertyItem` returned by the Notion API containing the `Checkbox` property."""


class DatePropertyItem(PropertyItem, Date, type='date'):
    """A `PropertyItem` returned by the Notion API containing the `Date` property."""


class StatusPropertyItem(PropertyItem, Status, type='status'):
    """A `PropertyItem` returned by the Notion API containing the `Status` property."""


class SelectPropertyItem(PropertyItem, Select, type='select'):
    """A `PropertyItem` returned by the Notion API containing the `Select` property."""


class MultiSelectPropertyItem(PropertyItem, MultiSelect, type='multi_select'):
    """A `PropertyItem` returned by the Notion API containing the `MultiSelect` property."""


class PeoplePropertyItem(PropertyItem, type='people'):
    """A `PropertyItem` returned by the Notion API containing the `People` property."""

    # According to the Notion API docs, this should be a list of people objects.
    # Nevertheless, the Notion API returns a single people object
    people: User


class URLPropertyItem(PropertyItem, URL, type='url'):
    """A `PropertyItem` returned by the Notion API containing the `URL` property."""


class EmailPropertyItem(PropertyItem, Email, type='email'):
    """A `PropertyItem` returned by the Notion API containing the `Email` property."""


class PhoneNumberPropertyItem(PropertyItem, PhoneNumber, type='phone_number'):
    """A `PropertyItem` returned by the Notion API containing the `PhoneNumber` property."""


class FilesPropertyItem(PropertyItem, Files, type='files'):
    """A `FilesPropertyItem` returned by the Notion API containing the `Files` property."""


class FormulaPropertyItem(PropertyItem, Formula, type='formula'):
    """A `PropertyItem` returned by the Notion API containing the `Formula` property."""


class RelationPropertyItem(PropertyItem, type='relation'):
    """A `PropertyItem` returned by the Notion API containing many `Relation` properties."""

    relation: ObjectRef


class RollupPropertyItem(PropertyItem, Rollup, type='rollup'):
    """A `PropertyItem` returned by the Notion API containing the `Rollup` property."""


class CreatedTimePropertyItem(PropertyItem, CreatedTime, type='created_time'):
    """A `PropertyItem` returned by the Notion API containing the `CreatedTime` property."""


class CreatedByPropertyItem(PropertyItem, CreatedBy, type='created_by'):
    """A `PropertyItem` returned by the Notion API containing the `CreatedBy` property."""


class LastEditedTimePropertyItem(PropertyItem, LastEditedTime, type='last_edited_time'):
    """A `PropertyItem` returned by the Notion API containing the `LastEditedTime` property."""


class LastEditedByPropertyItem(PropertyItem, LastEditedBy, type='last_edited_by'):
    """A `PropertyItem` returned by the Notion API containing the `LastEditedBy` property."""


class UniqueIDPropertyItem(PropertyItem, UniqueID, type='unique_id'):
    """A `PropertyItem` returned by the Notion API containing the `UniqueID` property."""


class VerificationPropertyItem(PropertyItem, Verification, type='verification'):
    """A `PropertyItem` returned by the Notion API containing the `Verification` property."""


class ButtonPropertyItem(PropertyItem, Button, type='button'):
    """A `PropertyItem` returned by the Notion API containing the `Button` property."""


PAGINATED_PROP_VALS = (RichText, Title, People, Relation, Rollup)
f"""Property values that are potentially paginated when exceeding {MAX_ITEMS_PER_PROPERTY} of
inline page/person mentions or items and need to be fetched seperately from the server.

Source: https://developers.notion.com/reference/retrieve-a-page#limits
"""



"""Functionality for working with Markdown in Notion."""

from __future__ import annotations

import re
from collections.abc import Callable, Iterator, Sequence
from typing import TYPE_CHECKING, TypeGuard

import mistune
import numpy as np
from mistune.directives import FencedDirective, TableOfContents
from numpy.typing import NDArray

from ultimate_notion.obj_api.core import raise_unset
from ultimate_notion.utils import rank

if TYPE_CHECKING:
    from ultimate_notion.rich_text import Math, Mention, RichTextBase


MD_STYLES = ('bold', 'italic', 'strikethrough', 'code', 'link')
"""Markdown styles supported by Notion."""
MD_STYLE_MAP = {
    'bold': '**',
    'italic': '*',
    'strikethrough': '~~',
    'code': '`',
}
"""Mapping from markdown style to markdown symbol."""


def md_spans(rich_texts: Sequence[RichTextBase]) -> NDArray[np.int_]:
    """Convert rich text to markdown spans.

    An span is a sequence of rich texts with the same markdown style expressed as a row in the returned array.
    The value k of the j-th array column corresponds to the length of the current span richt_texts[j-k:j].
    """
    spans: NDArray[np.int_] = np.zeros((len(MD_STYLES), len(rich_texts) + 1), dtype=int)
    old_ranks: NDArray[np.int_] = np.zeros(len(MD_STYLES), dtype=int)
    prev_rich_text = None

    for j, rich_text in enumerate(rich_texts, start=1):
        for i, md_style in enumerate(MD_STYLES):
            if md_style == 'link':
                if not rich_text.is_text:
                    continue
                href = rich_text.obj_ref.href
                if href is not None:
                    prev_href = prev_rich_text.obj_ref.href if prev_rich_text is not None else None
                    if href == prev_href:  # continue current link span or start new one
                        spans[i, j] = spans[i, j - 1] + 1
                    else:
                        spans[i, j] = 1
            else:
                annotations = rich_text.obj_ref.annotations
                if getattr(annotations, md_style) is True:
                    spans[i, j] = spans[i, j - 1] + 1

        # handle the case of overlapping spans, e.g. **abc ~~def** ghi~~ -> **abc ~~def~~** ~~ghi~~
        curr_ranks = rank(-spans[:, j])
        for i in np.where(curr_ranks < old_ranks)[0]:
            spans[i, j] = 1 if spans[i, j] > 0 else 0  # start a new span if an encompassing span ends
        old_ranks = curr_ranks
        prev_rich_text = rich_text

    return spans


def sorted_md_spans(md_spans: NDArray[np.int_]) -> Iterator[tuple[int, int, str]]:
    """Sort the spans of the given markdown spans in the right order.

    We have to iterate from the smallest spans to the largest spans and from left to right.
    """
    sorted_spans = []
    md_spans = md_spans.copy()
    for span_len in reversed(range(1, np.max(md_spans) + 1)):
        indices = tuple(idx[::-1] for idx in np.where(md_spans == span_len))
        for i, j in zip(*indices, strict=True):
            sorted_spans.append((j - span_len, j - 1, MD_STYLES[i]))
            md_spans[i, j - span_len + 1 : j + 1] = 0
    return reversed(sorted_spans)


def rich_texts_to_markdown(rich_texts: Sequence[RichTextBase]) -> str:
    """Convert a list of rich texts to markdown."""

    def has_only_ws_chars(text: str) -> bool:
        return re.match(r'^\s*$', text) is not None

    def first_non_ws_char(text: str) -> re.Match[str] | None:
        return re.search(r'\S', text)

    def last_non_ws_char(text: str) -> re.Match[str] | None:
        return re.search(r'\S(?=\s*$)', text)

    def add_md_style(
        md_rich_texts: list[str], rich_texts: list[RichTextBase], start: int, end: int, md_style: str
    ) -> None:
        # we skip text blocks with only whitespace characters
        if has_only_ws_chars(md_rich_texts[start]) and start != end:
            return add_md_style(md_rich_texts, rich_texts, start + 1, end, md_style)
        elif has_only_ws_chars(md_rich_texts[end]) and start != end:
            return add_md_style(md_rich_texts, rich_texts, start, end - 1, md_style)

        left = md_rich_texts[start]
        if lmatch := first_non_ws_char(left):
            if md_style == 'link':
                md_rich_texts[start] = left[: lmatch.start()] + '[' + left[lmatch.start() :]
            else:
                md_rich_texts[start] = left[: lmatch.start()] + MD_STYLE_MAP[md_style] + left[lmatch.start() :]

        right = md_rich_texts[end]
        if rmatch := last_non_ws_char(right):
            if md_style == 'link':
                link = rich_texts[end].obj_ref.href
                md_rich_texts[end] = right[: rmatch.end()] + f']({link})' + right[rmatch.end() :]
            else:
                md_rich_texts[end] = right[: rmatch.end()] + MD_STYLE_MAP[md_style] + right[rmatch.end() :]

        if bool(lmatch) != bool(rmatch):  # should never happen!
            rt_objs = '\n'.join(str(rt.obj_ref) for rt in rich_texts)
            msg = f'Error when inserting markdown styles into:\n{rt_objs}'
            raise ValueError(msg)

    def add_all_md_styles(md_rich_texts: list[str], rich_texts: list[RichTextBase]) -> None:
        for start, end, md_style in sorted_md_spans(md_spans(rich_texts)):
            add_md_style(md_rich_texts, rich_texts, start, end, md_style)

    def add_mentions(md_rich_texts: list[str], rich_texts: list[RichTextBase]) -> None:
        def is_mention(rt: RichTextBase) -> TypeGuard[Mention]:
            return rt.is_mention

        def is_math(rt: RichTextBase) -> TypeGuard[Math]:
            return rt.is_equation

        for idx, text in enumerate(rich_texts):
            if is_math(text):
                md_rich_texts[idx] = '$' + text.obj_ref.plain_text.strip() + '$'
            elif is_mention(text):
                obj_ref = text.obj_ref
                match text.type:
                    case 'user' | 'date':
                        md_rich_texts[idx] = f'[{obj_ref.plain_text}]()'  # @ is already included
                    case 'custom_emoji':
                        md_rich_texts[idx] = f'{obj_ref.plain_text}'  # parentheses are already included
                    case _:
                        md_rich_texts[idx] = f'↗[{obj_ref.plain_text}]({obj_ref.href})'

    def find_span(
        rich_texts: list[RichTextBase], style_cond: Callable[[RichTextBase], bool]
    ) -> Iterator[tuple[int, int]]:
        left: int | None = None
        for idx, text in enumerate(rich_texts):
            if style_cond(text) and left is None:
                left = idx
            elif not style_cond(text) and left is not None:
                yield left, idx - 1
                left = None
        if left is not None:
            yield left, len(rich_texts) - 1

    def add_underlines(md_rich_texts: list[str], rich_texts: list[RichTextBase]) -> None:
        for left, right in find_span(rich_texts, lambda rt: raise_unset(rt.obj_ref.annotations).underline):
            md_rich_texts[left] = '<u>' + md_rich_texts[left]
            md_rich_texts[right] += '</u>'

    rich_texts = list(rich_texts)
    md_rich_texts = [rich_text.obj_ref.plain_text for rich_text in rich_texts]
    add_mentions(md_rich_texts, rich_texts)
    add_all_md_styles(md_rich_texts, rich_texts)
    add_underlines(md_rich_texts, rich_texts)

    return ''.join(md_rich_texts)


def md_comment(text: str) -> str:
    """Create a markdown comment."""
    return f'<!--- {text} -->\n'


def get_md_renderer() -> Callable[[str], str]:
    """Create a markdown renderer."""

    vanilla_renderer = mistune.create_markdown(
        plugins=[
            'strikethrough',
            'url',
            'task_lists',
            'math',
            'table',
            FencedDirective(
                [
                    TableOfContents(),
                ]
            ),
        ],
        escape=False,
    )

    def md_renderer(md_str: str) -> str:
        match html := vanilla_renderer(md_str):
            case str():
                return html
            case list():
                msg = 'Cannot convert rich text to HTML, because the renderer returned a list:\n'
                msg += '\n'.join(str(elem) for elem in html)
                raise ValueError(msg)
            case _:
                msg = f'Cannot convert rich text to HTML, because the renderer returned {type(html)}'
                raise ValueError(msg)

    return md_renderer


render_md = get_md_renderer()
"""Convert Markdown to HTML."""
